
(******************************************************************************)
(*                         DRBD Operators                                     *)
(*                      Definitions and Probabilities                         *)
(*                       and Simplification Theorems                          *)
(*                                  by                                        *)
(*                          Yassmeen Elderhalli                               *)
(*                      y_elderh@encs.concordia.ca                            *)
(******************************************************************************)

(*loadPath := "/home/yassmeen/Downloads/Required-Theories" :: !loadPath;
loadPath := "/home/yassmeen/Downloads/DFT-Theories" :: !loadPath;

*)
(*set_trace "Unicode" 0;*)



(*app load ["HolKernel", "Parse", "boolLib", "bossLib", "numLib", "unwindLib", 
"tautLib", "Arith", "prim_recTheory", "combinTheory", "quotientTheory", 
"arithmeticTheory", "realTheory", "hrealTheory", "realaxTheory", "realLib",  
"jrhUtils", "pairTheory", "seqTheory", "limTheory", "transcTheory", "listTheory", 
"mesonLib", "boolTheory", "topologyTheory", "pred_setTheory", "util_probTheory", 
"optionTheory", "numTheory", "sumTheory", "InductiveDefinition", "ind_typeTheory",
"pred_setLib", "iterate_hvgTheory", "card_hvgTheory", "product_hvgTheory", 
"topology_hvgTheory", "derivative_hvgTheory", "integration_hvgTheory",
"real_sigmaTheory", "extreal_hvgTheory", "measure_hvgTheory", "lebesgue_hvgTheory",
"probability_hvgTheory", "lebesgue_measure_hvgTheory", "normal_rv_hvgTheory", "netsTheory",
"PIE_EXTREALTheory", "integration_before_afterTheory", "dep_rewrite", "DFT_Gates_def_PROBTheory", "WCSPTheory"];*)

open HolKernel Parse boolLib bossLib numLib unwindLib tautLib Arith prim_recTheory 
combinTheory quotientTheory arithmeticTheory hrealTheory realaxTheory realTheory 
realLib jrhUtils pairTheory seqTheory limTheory transcTheory listTheory mesonLib 
boolTheory topologyTheory pred_setTheory util_probTheory optionTheory numTheory 
sumTheory InductiveDefinition ind_typeTheory pred_setLib iterate_hvgTheory 
card_hvgTheory product_hvgTheory topology_hvgTheory derivative_hvgTheory
integration_hvgTheory real_sigmaTheory extreal_hvgTheory measure_hvgTheory
lebesgue_hvgTheory probability_hvgTheory lebesgue_measure_hvgTheory
normal_rv_hvgTheory netsTheory PIE_EXTREALTheory integration_before_afterTheory dep_rewrite DFT_Gates_def_PROBTheory  WCSPTheory;

val _ = new_theory "DRBD";



infixr 0 ++ << || ORELSEC ## --> THENC;
infix 1 >> |->;
fun parse_with_goal t (asms, g) =
  let
    val ctxt = free_varsl (g::asms)
  in
    Parse.parse_in_context ctxt t
  end;
 
val PARSE_TAC = fn tac => fn q => W (tac o parse_with_goal q);
val Suff = PARSE_TAC SUFF_TAC;
val POP_ORW = POP_ASSUM (fn thm => ONCE_REWRITE_TAC [thm]);
val !! = REPEAT;
val op++ = op THEN;
val op<< = op THENL;
val op|| = op ORELSE;
val op>> = op THEN1;
val std_ss' = simpLib.++ (std_ss, boolSimps.ETA_ss);

(*---------------------------*)
fun SET_TAC L =
    POP_ASSUM_LIST(K ALL_TAC) THEN REPEAT COND_CASES_TAC THEN
    REWRITE_TAC (append [EXTENSION, SUBSET_DEF, PSUBSET_DEF, DISJOINT_DEF,
    SING_DEF] L) THEN SIMP_TAC std_ss [NOT_IN_EMPTY, IN_UNIV, IN_UNION,
    IN_INTER, IN_DIFF, IN_INSERT, IN_DELETE, IN_BIGINTER, IN_BIGUNION,
    IN_IMAGE, GSPECIFICATION, IN_DEF] THEN METIS_TAC [];

fun SET_RULE tm = prove(tm,SET_TAC []);


val set_rewrs
= [INTER_EMPTY, INTER_UNIV, UNION_EMPTY, UNION_UNIV, DISJOINT_UNION,
DISJOINT_INSERT, DISJOINT_EMPTY, GSYM DISJOINT_EMPTY_REFL,
DISJOINT_BIGUNION, INTER_SUBSET, INTER_IDEMPOT, UNION_IDEMPOT,
SUBSET_UNION];

val UNIONL_def = Define `(UNIONL [] = {})
/\ (UNIONL (s::ss) = (s:'a ->bool) UNION UNIONL ss)`;


val IN_UNIONL = store_thm
("IN_UNIONL",
``!l (v:'a ). v IN UNIONL l = (?s. MEM s l /\ v IN s)``,
Induct >> RW_TAC std_ss [UNIONL_def, MEM, NOT_IN_EMPTY]
++ RW_TAC std_ss [UNIONL_def, MEM, NOT_IN_EMPTY, IN_UNION]
++ PROVE_TAC []);


val elt_rewrs
= [SUBSET_DEF, IN_COMPL, IN_DIFF, IN_UNIV, NOT_IN_EMPTY, IN_UNION,
IN_INTER, IN_IMAGE, IN_FUNSET, IN_DFUNSET, GSPECIFICATION,
DISJOINT_DEF, IN_BIGUNION, IN_BIGINTER, IN_COUNT, IN_o,
IN_UNIONL, IN_DELETE, IN_PREIMAGE, IN_SING, IN_INSERT];


fun rewr_ss ths =
simpLib.++
(std_ss,
simpLib.SSFRAG
{ac = [],
name = NONE,
convs = [],
dprocs = [],
filter = NONE,
rewrs = set_rewrs @ elt_rewrs,
congs = []});
val pset_set_ss = rewr_ss set_rewrs;
val pset_elt_ss = rewr_ss elt_rewrs;
val pset_ss = rewr_ss (set_rewrs @ elt_rewrs);


fun PSET_TAC ths =
REPEAT (POP_ASSUM MP_TAC)
++ RW_TAC pset_set_ss ths
++ REPEAT (POP_ASSUM MP_TAC)
++ RW_TAC pset_elt_ss ths;

fun parse_with_goal t (asms, g) =
  let
    val ctxt = free_varsl (g::asms)
  in
    Parse.parse_in_context ctxt t
  end;
val PARSE_TAC = fn tac => fn q => W (tac o parse_with_goal q);
val Simplify = RW_TAC arith_ss;
val Suff = PARSE_TAC SUFF_TAC;
val Know = PARSE_TAC KNOW_TAC;
fun wrap a = [a];
val Rewr = DISCH_THEN (REWRITE_TAC o wrap);
val Rewr' = DISCH_THEN (ONCE_REWRITE_TAC o wrap);
val Cond =
  DISCH_THEN
  (fn mp_th =>
   let
     val cond = fst (dest_imp (concl mp_th))
   in
     KNOW_TAC cond << [ALL_TAC, DISCH_THEN (MP_TAC o MP mp_th)]
   end);
val FUN_EQ_TAC = CONV_TAC (CHANGED_CONV (ONCE_DEPTH_CONV FUN_EQ_CONV));
fun K_TAC _ = ALL_TAC;
fun MESON ths tm = prove(tm,MESON_TAC ths);
fun METIS ths tm = prove(tm,METIS_TAC ths);
val DISC_RW_KILL = DISCH_TAC THEN ONCE_ASM_REWRITE_TAC [] THEN 
                   POP_ASSUM K_TAC;
			   
val IN_REST = store_thm ("IN_REST",
 ``!x:'a. !s. x IN (REST s) <=> x IN s /\ ~(x = CHOICE s)``,
  REWRITE_TAC[REST_DEF, IN_DELETE]);
fun SET_TAC L = 
    POP_ASSUM_LIST(K ALL_TAC) THEN REPEAT COND_CASES_TAC THEN
    REWRITE_TAC (append [EXTENSION, SUBSET_DEF, PSUBSET_DEF, DISJOINT_DEF,
    SING_DEF] L) THEN
    SIMP_TAC std_ss [NOT_IN_EMPTY, IN_UNIV, IN_UNION, IN_INTER, IN_DIFF, 
      IN_INSERT, IN_DELETE, IN_REST, IN_BIGINTER, IN_BIGUNION, IN_IMAGE, 
      GSPECIFICATION, IN_DEF, EXISTS_PROD] THEN METIS_TAC [];
fun ASSERT_TAC tm = SUBGOAL_THEN tm STRIP_ASSUME_TAC;
fun SET_RULE tm = prove(tm,SET_TAC []);
fun ASM_SET_TAC L = REPEAT (POP_ASSUM MP_TAC) THEN SET_TAC L;
val ASM_ARITH_TAC = REPEAT (POP_ASSUM MP_TAC) THEN ARITH_TAC;
val ASM_REAL_ARITH_TAC = REPEAT (POP_ASSUM MP_TAC) THEN REAL_ARITH_TAC;




(******************************************************************************)
(*                                                                            *)
(*                     DRBD Operators and constructs  Definitions             *)
(*                                                                            *)
(******************************************************************************)

val DRBD_event_def = Define ` 
          DRBD_event  p (X:'a->extreal) t  =  {s| Normal t < X s} INTER p_space p`;
val Rel_def = Define `Rel p X t = prob p (DRBD_event p X t) `;


val R_OR_def = Define `R_OR (X: 'a->extreal) Y   = (\s. max  (X s) (Y s))`;

val R_AND_def = Define `R_AND (X: 'a->extreal) Y   = (\s. min (X s)( Y s))`;

val R_SIMULT_def = Define ` 
     R_SIMULT (X: 'a->extreal) Y   = (\s. if  ((X s) = ( Y s )) then (X s) else PosInf)`;

val R_BEFORE_def = Define `
     R_BEFORE (X: 'a->extreal) Y   = (\s. if  (X s) < (Y s) then (X s) else PosInf)`;

val R_AFTER_def = Define `
     R_AFTER (X: 'a->extreal) Y   = (\s. if  (Y s) < (X s) then (X s) else PosInf)`;


val R_INCLUSIVE_BEFORE_def = Define ` 
     R_INCLUSIVE_BEFORE (X:'a->extreal) Y   = (\s. if  (X s) <=(Y s) then (X s) else PosInf)`;

val R_INCLUSIVE_AFTER_def = Define ` 
     R_INCLUSIVE_AFTER (X:'a->extreal) Y   = (\s. if  (Y s) <=(X s) then (X s) else PosInf)`;


val R_NEVER_def = Define ` R_NEVER = (\s. PosInf)`;

val R_ALWAYS_def = Define ` R_ALWAYS = (\s. 0:extreal)`;


val R_CSP_def = Define `R_CSP (Y:'a-> extreal) X =( \s. if (Y s < X s) then  X s else PosInf)`;

val R_WSP_def = Define ` 
    R_WSP (Y:'a->extreal) X_a X_d = 
    R_AND (R_AFTER X_a Y)(R_AFTER Y X_d)`;

val R_HSP_def = Define ` R_HSP (Y:'a->extreal) X  =(\s.  max (Y s )  (X s)) `;



(* ========================================================================== *)
(*                                                                            *)
(*                      verified equivalent definitions                       *)
(*                                                                            *)
(* ========================================================================== *)


val HSP_OR = store_thm("HSP_OR",
  ``  ! X Y. R_HSP Y X  = R_OR Y X``,
  RW_TAC std_ss [R_HSP_def, R_OR_def]);


val R_WSP_def1 = store_thm("R_WSP_def1",
  `` !X_a X_d Y. R_WSP Y X_a X_d =  R_AND (R_AFTER Y X_d) (R_AFTER X_a Y)  ``,
    RW_TAC std_ss[R_WSP_def, R_AND_def]
    THEN RW_TAC std_ss[min_comm]);

val RWSP_RHSP = store_thm("RWSP_RHSP",
  `` !X Y. (!s. ALL_DISTINCT [ Y s; X s]) ==> (R_WSP Y X X = R_HSP Y X) ``,
  RW_TAC std_ss [R_WSP_def1, R_HSP_def, R_OR_def, R_AND_def, R_AFTER_def]
  THEN KNOW_TAC(``
      !s.(
        min (if X s < Y s then Y s else PosInf)
          (if Y s < X s then X s else PosInf)) =
         (max (Y s) (X s))``)

       THEN1(RW_TAC std_ss[extreal_max_def, extreal_min_def] THEN REPEAT( POP_ASSUM MP_TAC)
THEN  RW_TAC std_ss[] THEN METIS_TAC [le_lt, lt_le, lt_infty, le_infty, extreal_lt_def, extreal_le_def, lte_trans, let_trans, ALL_DISTINCT, MEM])
THEN RW_TAC std_ss[]);


val RWSP_RCSP = store_thm("RWSP_RCSP",

`` !X_a X_d Y.
            (!s. ALL_DISTINCT [X_a s; Y s]) /\ (X_d = R_NEVER) ==>
            (R_WSP Y X_a X_d = R_CSP Y X_a)``,

   RW_TAC std_ss[R_WSP_def1, R_CSP_def, R_OR_def, R_AND_def, R_AFTER_def, ALL_DISTINCT, MEM, R_NEVER_def]
     THEN KNOW_TAC(``
      !s.(
        min (if PosInf < Y s then Y s else PosInf)
          (if Y s < X_a s then X_a s else PosInf)) =
        (if Y s < X_a s then X_a s else PosInf)``)
          THEN1(RW_TAC std_ss[extreal_max_def, extreal_min_def, R_NEVER_def] THEN REPEAT( POP_ASSUM MP_TAC)
THEN  RW_TAC std_ss[] THEN FULL_SIMP_TAC std_ss[ALL_DISTINCT, MEM] THEN METIS_TAC [le_lt, lt_le, lt_infty, le_infty, extreal_lt_def, extreal_le_def, lte_trans, let_trans, ALL_DISTINCT, MEM])
THEN RW_TAC std_ss[]);


(* -------------------------------------------------------------------------- *)
(*                                                                            *)
(*                      Simplification Theorems                               *)
(*                                                                            *)
(* -------------------------------------------------------------------------- *)


val R_AND_ID = store_thm("R_AND_ID",
  `` !  X . R_AND X X = X``,
RW_TAC std_ss [R_AND_def] THEN
RW_TAC std_ss [min_refl] THEN
RW_TAC std_ss [ETA_THM]);


val R_AND_COMM = store_thm("R_AND_COMM",
  `` ! X Y . R_AND X Y = R_AND Y X ``,
REPEAT GEN_TAC
THEN REWRITE_TAC [R_AND_def]
THEN METIS_TAC[min_comm]);

val R_AND_ASSOC = store_thm("R_AND_ASSOC",
  `` ! X Y Z . R_AND (R_AND X Y) Z = R_AND X (R_AND Y Z)  ``,
REPEAT GEN_TAC
THEN REWRITE_TAC [R_AND_def]
THEN RW_TAC std_ss [extreal_min_def]
THEN KNOW_TAC (``
      !s.
         (if (if (X :'a -> extreal) s <= (Y :'a -> extreal) s then X s
         else Y s) <= (Z :'a -> extreal) s then
           if X s <= Y s then X s else Y s
          else Z s) =
	 (if X s <= if Y s <= Z s then Y s else Z s then X s
          else if Y s <= Z s then Y s
          else Z s)``)
     THEN1( RW_TAC std_ss[]
            THEN REPEAT(POP_ASSUM MP_TAC)
	    THEN RW_TAC std_ss []
	    THEN FULL_SIMP_TAC std_ss[]
	    THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def, extreal_not_infty])
THEN RW_TAC std_ss[]);

val R_OR_ID = store_thm("R_OR_ID",
  `` ! X . R_OR X X  =X``,

RW_TAC std_ss [R_OR_def]
THEN RW_TAC std_ss[max_refl]
THEN RW_TAC std_ss [ETA_THM]);


val R_AND_NEVER = store_thm("R_AND_NEVER",
  `` ! X . R_AND X R_NEVER = X ``,
RW_TAC std_ss [R_AND_def, R_NEVER_def, min_infty]
THEN RW_TAC std_ss [ETA_THM]);


val R_OR_ALWAYS = store_thm("R_OR_ALWAYS",
  ``! X . (!s. 0 <=X s) ==> (R_OR X R_ALWAYS = X)``,
REPEAT GEN_TAC
THEN REWRITE_TAC [R_OR_def, R_ALWAYS_def]
THEN RW_TAC std_ss [extreal_max_def]
THEN KNOW_TAC (``!s.  ((if (X :'a -> extreal) s <= (0 :extreal) then (0 :extreal) else X s)=X s)``)      THEN1(RW_TAC std_ss[]
          THEN METIS_TAC[lt_le, lte_trans, let_trans])
THEN RW_TAC std_ss [ETA_THM]);


val R_OR_COMM = store_thm("R_OR_COMM",
  ``!  X Y . R_OR X Y    = R_OR Y X ``,
RW_TAC std_ss [R_OR_def]
THEN RW_TAC std_ss [max_comm]);


val R_OR_ASSOC = store_thm("R_OR_ASSOC",
  `` !X Y Z .   (R_OR ( R_OR X Y ) Z  =   R_OR X (R_OR Y Z ) ) ``,

RW_TAC std_ss [R_OR_def]
THEN RW_TAC std_ss [extreal_max_def]
THEN KNOW_TAC (``
      !s. (if (if (X :'a -> extreal) s <= (Y :'a -> extreal) s then Y s
            else X s) <= (Z :'a -> extreal) s then Z s
           else if X s <= Y s then Y s
           else X s) =
	  (if X s <= if Y s <= Z s then Z s else Y s then
            if Y s <= Z s then Z s else Y s
           else X s)  ``)
      THEN1( RW_TAC std_ss[]
            THEN REPEAT(POP_ASSUM MP_TAC)
	    THEN RW_TAC std_ss []
	    THEN FULL_SIMP_TAC std_ss[]
	    THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def, extreal_not_infty])
THEN RW_TAC std_ss[]);


val R_OR_NEVER = store_thm("R_OR_NEVER",
  `` ! X . R_OR X R_NEVER = R_NEVER``,

RW_TAC std_ss [R_OR_def, max_infty, R_NEVER_def]);


val  R_OR_OVER_AND = store_thm("R_OR_OVER_AND",
  ``! X Y Z . R_OR X (R_AND Y Z ) = R_AND (R_OR X Y ) ( R_OR X Z)``,
RW_TAC std_ss [R_AND_def, R_OR_def]
THEN KNOW_TAC (``
      !s.
          max ((X :'a -> extreal) s)(min ((Y :'a -> extreal) s) ((Z :'a -> extreal) s)) =
	  min (max (X s) (Y s)) (max (X s) (Z s))``)
    THEN1(RW_TAC std_ss[extreal_min_def, extreal_max_def]
          THEN REPEAT(POP_ASSUM MP_TAC)
  	  THEN RW_TAC std_ss []
	  THEN FULL_SIMP_TAC std_ss[]
	  THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def, extreal_not_infty])
THEN RW_TAC std_ss []);


val R_AND_OVER_OR = store_thm("R_AND_OVER_OR",
  ``! X Y Z . R_AND X (R_OR Y Z ) = R_OR(R_AND X Y ) ( R_AND X Z)``,
RW_TAC std_ss [R_AND_def, R_OR_def]
THEN KNOW_TAC (``
     !s.
       min ((X :'a -> extreal) s)(max ((Y :'a -> extreal) s) ((Z :'a -> extreal) s)) =
       max (min (X s) (Y s)) (min (X s) (Z s)) ``)
     THEN1 (RW_TAC std_ss[extreal_min_def, extreal_max_def]
           THEN REPEAT(POP_ASSUM MP_TAC)
 	   THEN RW_TAC std_ss []
	   THEN FULL_SIMP_TAC std_ss[]
	   THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def, extreal_not_infty])
THEN RW_TAC std_ss []);

val R_AND_ALWAYS = store_thm("R_AND_ALWAYS",
  `` ! X. (!s. 0 <= X s)==> (R_AND X R_ALWAYS = R_ALWAYS)``,
RW_TAC std_ss [R_AND_def, R_ALWAYS_def]
THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);


val R_ABSORB_OR = store_thm("R_ABSORB_OR",
  ``  ! X Y . R_AND X ( R_OR X Y ) = X``,
RW_TAC std_ss [R_OR_def, R_AND_def]
THEN RW_TAC std_ss [extreal_min_def, extreal_max_def]
THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);


val R_ABSORB_OR = store_thm("R_ABSORB_OR",
  ``! X Y . R_OR X (R_AND X Y ) =X``,
RW_TAC std_ss [R_OR_def, R_AND_def]
THEN RW_TAC std_ss [extreal_min_def, extreal_max_def]
THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);


val R_INCLUSIVE_AFTER_def2 = store_thm("R_INCLUSIVE_AFTER_def2",
  ``! X Y . R_INCLUSIVE_AFTER X Y = R_AND (R_AFTER X Y ) (R_SIMULT X Y)``,
REPEAT GEN_TAC
THEN REWRITE_TAC [R_AND_def, R_AFTER_def, R_SIMULT_def, R_INCLUSIVE_AFTER_def]
THEN KNOW_TAC(``!s. 
    (if Y s <= X s then X s else PosInf) =
    (min ((\s. if Y s < X s then X s else PosInf) s)
          ((\s. if X s = Y s then X s else PosInf) s))``)
    THEN1( RW_TAC std_ss[extreal_min_def]
            THEN REPEAT(POP_ASSUM MP_TAC)
	    THEN RW_TAC std_ss []
	    THEN FULL_SIMP_TAC std_ss[]
	    THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def, extreal_not_infty])
THEN RW_TAC std_ss[]);

val R_AFTER_NON_COMM = store_thm("R_AFTER_NON_COMM",
  `` ! X Y . R_OR (R_AFTER X Y)(R_AFTER Y X) = R_NEVER``,

REPEAT GEN_TAC
THEN REWRITE_TAC [R_OR_def, R_AND_def, R_AFTER_def, R_NEVER_def]
THEN RW_TAC std_ss[]
THEN METIS_TAC [ min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);

val R_SIMULT_COMM = store_thm("R_SIMULT_COMM",
  `` !X Y. R_SIMULT X  Y = R_SIMULT Y X ``,
METIS_TAC[R_SIMULT_def]);

val R_AFTER_OVER_AND = store_thm("R_AFTER_OVER_AND",
  ``!X Y Z. R_AFTER X ( R_AND Y Z) = R_AND (R_AFTER X Y) (R_AFTER X Z)``,
RW_TAC std_ss[R_AFTER_def,R_AND_def , R_OR_def ]
THEN KNOW_TAC(``!s.
          (if min (Y s) (Z s) < X s then X s else PosInf) =
	  (min (if Y s < X s then X s else PosInf)
            (if Z s < X s then X s else PosInf)) ``)
     THEN1( RW_TAC std_ss[extreal_min_def]
            THEN REPEAT(POP_ASSUM MP_TAC)
	    THEN RW_TAC std_ss []
	    THEN FULL_SIMP_TAC std_ss[]
	    THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def, extreal_not_infty])
THEN RW_TAC std_ss[]);

val R_LEFT_AFTER_OVER_OR = store_thm("R_LEFT_AFTER_OVER_OR",
  ``! X Y Z. R_AFTER X (R_OR Y Z )  = R_OR (R_AFTER X Y)(R_AFTER X Z)``,
RW_TAC std_ss[R_AFTER_def,R_AND_def , R_OR_def ]
THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def]);

val R_LEFT_AFTER_OVER_OR = store_thm("R_LEFT_AFTER_OVER_OR",
  `` ! X Y Z. R_AFTER  X (R_OR Y Z )  = R_OR(R_AFTER  X Y)(R_AFTER  X Z)``,
RW_TAC std_ss[R_AFTER_def,R_AND_def , R_OR_def ]
THEN KNOW_TAC(``  
   !s.
      ( if max (Y s) (Z s) < X s then X s else PosInf) =
      (max (if Y s < X s then X s else PosInf)
          (if Z s < X s then X s else PosInf)) ``)

     THEN1(RW_TAC std_ss[extreal_max_def, extreal_min_def]
           THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def])
THEN RW_TAC std_ss[]);	   

val R_AFTER_AFTER = store_thm("R_AFTER_AFTER",
  `` ! X Y Z. R_AFTER X (R_AFTER Y Z) = R_OR(R_OR (R_AFTER X Y)(R_AFTER X Z))(R_AFTER Y Z) ``,
RW_TAC std_ss[R_AFTER_def, R_OR_def ]
THEN KNOW_TAC(``!s. (if (if Z s < Y s then Y s else PosInf) < X s then X s else PosInf) =
                     (max (max (if Y s < X s then X s else PosInf)
                      (if Z s < X s then X s else PosInf))
                      (if Z s < Y s then Y s else PosInf))``)
       THEN1(RW_TAC std_ss[]
            THEN METIS_TAC [min_infty, max_infty, lt_infty, lt_le, lt_total, lte_trans, let_trans, extreal_lt_def, extreal_le_def, extreal_min_def, extreal_max_def])
THEN RW_TAC std_ss[]);	    



(******************************************************************************)
(*                                                                            *)
(*                     some required properties of extreal and real           *)
(*                                                                            *)
(******************************************************************************)


val normal_real_mul1 = store_thm("normal_real_mul1",
  ``  !a b c.
            a <> PosInf /\ a <> NegInf /\ b <> PosInf /\ b <> NegInf ==>
            (Normal (real a * real b * c) = a * b * Normal c)``,
REPEAT Cases
THEN RW_TAC std_ss[]
THEN RW_TAC std_ss[real_normal]
THEN RW_TAC std_ss[extreal_mul_def]);


val normal_real_mul2 = store_thm("normal_real_mul2",
  ``    !a b c d.
            a <> PosInf /\ a <> NegInf /\ b <> PosInf /\ b <> NegInf ==>
            (Normal (real a * c * d * real b) = a * b * Normal (c * d)) ``,
REPEAT Cases
THEN RW_TAC std_ss[]
THEN RW_TAC std_ss[real_normal]
THEN RW_TAC std_ss[extreal_mul_def]
THEN METIS_TAC[REAL_MUL_ASSOC, REAL_MUL_COMM]);


val normal_real_mul3 = store_thm("normal_real_mul3",
  `` !a b c.
            a <> PosInf /\ a <> NegInf /\ b <> PosInf /\ b <> NegInf ==>
            (Normal (real a * c * real b) = a * b * Normal c)``,
REPEAT Cases
THEN RW_TAC std_ss[]
THEN RW_TAC std_ss[real_normal]
THEN RW_TAC std_ss[extreal_mul_def]
THEN METIS_TAC[REAL_MUL_ASSOC, REAL_MUL_COMM]);

val real_mul_real = store_thm("real_mul_real",
  ``!a b.
        a <> PosInf /\ a <> NegInf /\ b <> PosInf /\ b<> NegInf ==>
        (real a * real b = real (a * b))  ``,
RW_TAC std_ss[real_def]
THEN RW_TAC std_ss[REAL_MUL_LZERO] 
THEN SELECT_ELIM_TAC
THEN RW_TAC std_ss[]
     THEN1(Q.EXISTS_TAC`real a`
           THEN METIS_TAC[normal_real])
     THEN1(SELECT_ELIM_TAC
           THEN RW_TAC std_ss[]
                THEN1 (Q.EXISTS_TAC`real b`
                       THEN METIS_TAC[normal_real])
           THEN FULL_SIMP_TAC std_ss[extreal_mul_def]      
           THEN METIS_TAC[extreal_not_infty])
     THEN1(Q.EXISTS_TAC`real a`
           THEN METIS_TAC[normal_real])          
     THEN1(SELECT_ELIM_TAC
           THEN RW_TAC std_ss[]
                THEN1 (Q.EXISTS_TAC`real b`
                       THEN METIS_TAC[normal_real])
           THEN FULL_SIMP_TAC std_ss[extreal_mul_def]      
           THEN METIS_TAC[extreal_not_infty])
    THEN1(Q.EXISTS_TAC`real a`
           THEN METIS_TAC[normal_real])          
THEN SELECT_ELIM_TAC
THEN RW_TAC std_ss[]
     THEN1 (Q.EXISTS_TAC`real b`
            THEN METIS_TAC[normal_real])
THEN SELECT_ELIM_TAC
THEN RW_TAC std_ss[]
      THEN1(RW_TAC std_ss[extreal_mul_def])
THEN FULL_SIMP_TAC std_ss[extreal_mul_def]
THEN RW_TAC std_ss[] );
    	   

val extreal_sub_sub2 = store_thm("extreal_sub_sub2",
  ``   !a b.
            a <> PosInf /\ a <> NegInf /\ b <> PosInf /\ b <> NegInf ==>
            (a - (a - b) = b)  ``,
REPEAT Cases
THEN RW_TAC std_ss[]
THEN RW_TAC std_ss[extreal_sub_def]
THEN METIS_TAC[REAL_SUB_SUB2]);

val real_normal_real = store_thm("real_normal_real",
  `` !a . real (1-Normal(real(1 - a))) = 1 - real(1-a) ``,

RW_TAC std_ss[]
THEN `1 = Normal 1` by METIS_TAC[extreal_of_num_def]
THEN ONCE_ASM_REWRITE_TAC[]
THEN ONCE_REWRITE_TAC[extreal_sub_def]
THEN REWRITE_TAC [real_normal]);


(* -------------------------------------------------------------------------- *)
(*                                                                            *)
(*                Probability of Constructs and Operators                     *)
(*                                                                            *)
(* -------------------------------------------------------------------------- *)



val lborel_ge = store_thm("lborel_ge",
  `` ! x:real. {w|x <=w} IN measurable_sets lborel``,
RW_TAC std_ss[]
THEN RW_TAC std_ss [borel_closed, lborel, measurable_sets_def, subsets_def, CLOSED_INTERVAL_RIGHT ]);


val lborel_gt = store_thm("lborel_gt",
  `` ! x:real. {w|x <w} IN measurable_sets lborel``,
RW_TAC std_ss[]
THEN RW_TAC std_ss [borel_open, lborel, measurable_sets_def, subsets_def, OPEN_INTERVAL_RIGHT ]);

val PREIMAGE_EXTREAL_REAL_le = store_thm("PREIMAGE_EXTREAL_REAL_le",
  ``   
        !X t p.
            (!s. X s <> PosInf /\ 0 <= X s) ==>
            (PREIMAGE X {u | Normal t <= u} INTER p_space p =
             PREIMAGE (\s. real (X s)) {u | t <= u} INTER p_space p)``, 
RW_TAC std_ss []
THEN SRW_TAC [] [GSPECIFICATION, EXTENSION, PREIMAGE_def]
THEN EQ_TAC THEN1(RW_TAC std_ss[]
     	    THEN RW_TAC std_ss [real_def]
	    	 THEN1( METIS_TAC [lt_le, le_lt, lt_infty, le_infty, extreal_lt_def, extreal_le_def, extreal_of_num_def])
            THEN SELECT_ELIM_TAC
	    THEN RW_TAC std_ss[]
	    	 THEN1(Q.EXISTS_TAC` real (X x)`
		       THEN METIS_TAC [normal_real])
		 THEN METIS_TAC [lt_le, le_lt, lt_infty, le_infty, extreal_lt_def, extreal_le_def, extreal_of_num_def])
THEN RW_TAC std_ss[]
THEN FULL_SIMP_TAC std_ss [real_def]
THEN UNDISCH_TAC(`` t <= (if (X x = NegInf) \/ (X x = PosInf) then 0
                    else (@r. X x = Normal r))``)
THEN SELECT_ELIM_TAC
THEN RW_TAC std_ss[] THEN1 (Q.EXISTS_TAC` real (X x)`
     	    	            THEN METIS_TAC [normal_real,lt_le, le_lt, lt_infty, le_infty, extreal_lt_def, extreal_le_def, extreal_of_num_def])
THEN METIS_TAC [normal_real,lt_le, le_lt, lt_infty, le_infty, extreal_lt_def, extreal_le_def, extreal_of_num_def]);



val PREIMAGE_EXTREAL_REAL_lt = store_thm("PREIMAGE_EXTREAL_REAL_lt",
  ``   !X t p.
            (!s. X s <> PosInf /\ 0 <= X s) ==>
            (PREIMAGE X {u | Normal t < u} INTER p_space p =
             PREIMAGE (\s. real (X s)) {u | t < u} INTER p_space p)``, 
RW_TAC std_ss []
THEN SRW_TAC [] [GSPECIFICATION, EXTENSION, PREIMAGE_def]
THEN EQ_TAC THEN1(RW_TAC std_ss[]
     	    THEN RW_TAC std_ss [real_def]
	    	 THEN1( METIS_TAC [lt_le, le_lt, lt_infty, le_infty, extreal_lt_def, extreal_le_def, extreal_of_num_def])
            THEN SELECT_ELIM_TAC
	    THEN RW_TAC std_ss[]
	    	 THEN1(Q.EXISTS_TAC` real (X x)`
		       THEN METIS_TAC [normal_real])
		 THEN METIS_TAC [lt_le, le_lt, lt_infty, le_infty, extreal_lt_def, extreal_le_def, extreal_of_num_def, extreal_lt_eq])
THEN RW_TAC std_ss[]
THEN FULL_SIMP_TAC std_ss [real_def]
THEN UNDISCH_TAC(`` t< (if (X x = NegInf) \/ (X x = PosInf) then 0
                    else (@r. X x = Normal r))``)
THEN SELECT_ELIM_TAC
THEN RW_TAC std_ss[] THEN1 (Q.EXISTS_TAC` real (X x)`
     	    	            THEN METIS_TAC [normal_real,lt_le, le_lt, lt_infty, le_infty, extreal_lt_def, extreal_le_def, extreal_of_num_def])
THEN METIS_TAC [normal_real,lt_le, le_lt, lt_infty, le_infty, extreal_lt_def, extreal_le_def, extreal_of_num_def, extreal_lt_eq]);


val PREIMAGE_EXTREAL_REAL_lt1 = store_thm("PREIMAGE_EXTREAL_REAL_lt1",
  ``  !X t p.
            (!s. X s <> PosInf /\ 0 <= X s) ==>
            (PREIMAGE X {u | u < Normal t} INTER p_space p =
             PREIMAGE (\s. real (X s)) {u | u < t} INTER p_space p)``, 
RW_TAC std_ss []
THEN SRW_TAC [] [GSPECIFICATION, EXTENSION, PREIMAGE_def]
THEN EQ_TAC THEN1(RW_TAC std_ss[]
     	    THEN RW_TAC std_ss [real_def]
	    	 THEN1( METIS_TAC [lt_le, le_lt, lt_infty, le_infty, extreal_lt_def, extreal_le_def, extreal_of_num_def])
            THEN SELECT_ELIM_TAC
	    THEN RW_TAC std_ss[]
	    	 THEN1(Q.EXISTS_TAC` real (X x)`
		       THEN METIS_TAC [normal_real])
		 THEN METIS_TAC [lt_le, le_lt, lt_infty, le_infty, extreal_lt_def, extreal_le_def, extreal_of_num_def, extreal_lt_eq])
THEN RW_TAC std_ss[]
THEN FULL_SIMP_TAC std_ss [real_def]
THEN UNDISCH_TAC(`` (if (X x = NegInf) \/ (X x = PosInf) then 0
                     else (@r. X x = Normal r)) <t``)
THEN SELECT_ELIM_TAC
THEN RW_TAC std_ss[] THEN1 (Q.EXISTS_TAC` real (X x)`
     	    	            THEN METIS_TAC [normal_real,lt_le, le_lt, lt_infty, le_infty, extreal_lt_def, extreal_le_def, extreal_of_num_def])
THEN METIS_TAC [normal_real,lt_le, le_lt, lt_infty, le_infty, extreal_lt_def, extreal_le_def, extreal_of_num_def, extreal_lt_eq]);


val DRBD_event_in_event_p = store_thm("DRBD_event_in_event_p",
  ``    !X p t.
            rv_gt0_ninfinity [X] /\ random_variable (\s. real (X s)) p borel ==>
            DRBD_event p X t IN events p``,

RW_TAC std_ss[DRBD_event_def, rv_gt0_ninfinity_def]
THEN ` {s | Normal t < X s} INTER p_space p =
       PREIMAGE X {s | Normal t < s} INTER p_space p` by ASM_SET_TAC[PREIMAGE_def]
THEN RW_TAC std_ss[PREIMAGE_EXTREAL_REAL_lt]
THEN METIS_TAC[random_variable_def,borel_open, OPEN_INTERVAL_RIGHT, p_space_def, m_space_def, space_def, events_def, measurable_sets_def, subsets_def, IN_MEASURABLE]);


val PROB_DRBD_DFT_COMP = store_thm("PROB_DRBD_DFT_COMP",
  `` !X p t.
            rv_gt0_ninfinity [X] /\ random_variable (\s. real (X s)) p borel ==>
            (prob p (DRBD_event p X t) = 1 - prob p (DFT_event p X t)) ``,

RW_TAC std_ss[]
THEN ` DRBD_event p X t IN events p` by RW_TAC std_ss[DRBD_event_in_event_p]
THEN `prob_space p`by FULL_SIMP_TAC std_ss[random_variable_def]
THEN `prob p (p_space p DIFF (DRBD_event p X t)) = 1-prob p (DRBD_event p X t) ` by RW_TAC std_ss[PROB_COMPL]
THEN KNOW_TAC(`` prob p (DRBD_event p X t) = 1-prob p (p_space p DIFF DRBD_event p X t) ``)
     THEN1(RW_TAC std_ss[eq_sub_ladd, PROB_FINITE, EVENTS_COMPL]
          THEN RW_TAC std_ss[PROB_FINITE, sub_add2])
THEN DISCH_TAC
THEN KNOW_TAC(``p_space p DIFF DRBD_event p X t = DFT_event p X t ``)
    THEN1( RW_TAC std_ss[DRBD_event_def, DFT_event_def]
           THEN SRW_TAC  [][GSPECIFICATION, EXTENSION]
	   THEN METIS_TAC[extreal_lt_def, extreal_le_def, lt_le, le_lt, lt_trans, le_trans, lt_total, le_total])
THEN METIS_TAC[]);	   
	  

val DFT_event_in_event_p = store_thm("DFT_event_in_event_p",
  `` !X p t. rv_gt0_ninfinity[X] /\  random_variable (\s. real (X s)) p borel ==>
             DFT_event p X t IN events p``,

RW_TAC std_ss[DFT_event_def, rv_gt0_ninfinity_def]
THEN ` {s | X s <= Normal t} INTER p_space p =
       PREIMAGE X {s | s <= Normal t} INTER p_space p` by ASM_SET_TAC[PREIMAGE_def]
THEN RW_TAC std_ss[PREIMAGE_EXTREAL_REAL]

THEN METIS_TAC[random_variable_def,borel_closed, CLOSED_INTERVAL_LEFT, p_space_def, m_space_def, space_def, events_def, measurable_sets_def, subsets_def, IN_MEASURABLE]);


val PROB_DFT_DRBD_COMP = store_thm("PROB_DFT_DRBD_COMP",
  `` !X p t.
            rv_gt0_ninfinity [X] /\ random_variable (\s. real (X s)) p borel ==>
            (prob p (DFT_event p X t) = 1 - prob p (DRBD_event p X t)) ``,

RW_TAC std_ss[]
THEN ` DFT_event p X t IN events p` by RW_TAC std_ss[DFT_event_in_event_p]
THEN `prob_space p`by FULL_SIMP_TAC std_ss[random_variable_def]
THEN `prob p (p_space p DIFF (DFT_event p X t)) = 1-prob p (DFT_event p X t) ` by RW_TAC std_ss[PROB_COMPL]
THEN KNOW_TAC(`` prob p (DFT_event p X t) = 1-prob p (p_space p DIFF DFT_event p X t) ``)
     THEN1(RW_TAC std_ss[eq_sub_ladd, PROB_FINITE, EVENTS_COMPL]
          THEN RW_TAC std_ss[PROB_FINITE, sub_add2])
THEN DISCH_TAC
THEN KNOW_TAC(``p_space p DIFF DFT_event p X t = DRBD_event p X t ``)
    THEN1( RW_TAC std_ss[DRBD_event_def, DFT_event_def]
           THEN SRW_TAC  [][GSPECIFICATION, EXTENSION]
	   THEN METIS_TAC[extreal_lt_def, extreal_le_def, lt_le, le_lt, lt_trans, le_trans, lt_total, le_total])
THEN METIS_TAC[]);	   
	  


val Rel_CDF = store_thm("Rel_CDF",
  ``    !p X t.
            rv_gt0_ninfinity [X] /\ random_variable (\s. real (X s)) p borel ==>
            (Rel p X t = 1 - CDF p (\s. real (X s)) t)``,

RW_TAC std_ss[Rel_def, PROB_DRBD_DFT_COMP]
THEN METIS_TAC[CDF_prob_DFT_event, rv_gt0_ninfinity_def]);



(* ========================================================================== *)
(*              Relibaility of AND and OR operator                            *)
(* ========================================================================== *)

val R_AND_INTER = store_thm("R_AND_INTER",
  ``  !p t X Y.
            DRBD_event p (R_AND X Y) t =
            DRBD_event p X t INTER DRBD_event p Y t``,
RW_TAC std_ss [DRBD_event_def, R_AND_def] 
THEN SRW_TAC [] [GSPECIFICATION, EXTENSION]
THEN RW_TAC std_ss[extreal_min_def]
THEN EQ_TAC THEN RW_TAC std_ss[]
THEN METIS_TAC [extreal_lt_def, extreal_le_def, le_lt, lt_le, lt_trans, le_trans]);


val CDF_Rel = store_thm("CDF_Rel",
  ``    !p X t.
            rv_gt0_ninfinity [X] /\ random_variable (real o X ) p borel ==>
            (Rel p X t = 1 - CDF p (real o X) t)``,
RW_TAC std_ss[Rel_CDF, o_DEF]);


val R_OR_UNION = store_thm("R_OR_UNION",
  ``    !p t X Y.
            DRBD_event p (R_OR X Y) t =
            DRBD_event p X t UNION DRBD_event p Y t``,
RW_TAC std_ss [DRBD_event_def, R_OR_def] 
THEN SRW_TAC [] [GSPECIFICATION, EXTENSION]
THEN RW_TAC std_ss[extreal_max_def]
THEN EQ_TAC THEN RW_TAC std_ss[]
THEN METIS_TAC [extreal_lt_def, extreal_le_def, le_lt, lt_le, lt_trans, le_trans]);

val CDF_def3 = store_thm("CDF_def3",
  ``   !p X x.
            cont_CDF p X /\ random_variable X p borel ==>
            (CDF p X x = prob p (PREIMAGE X {y | y < x} INTER p_space p))``,

RW_TAC std_ss[cont_CDF_def]
THEN DEP_REWRITE_TAC[CDF_def2]
THEN `prob_space p` by FULL_SIMP_TAC std_ss[random_variable_def]
THEN KNOW_TAC(``!n. {y | x - 1 / &SUC n < y /\ y <= x}  IN subsets borel``)
     THEN1(RW_TAC std_ss[]
           THEN `{y | x - 1 / &SUC n < y /\ y <= x} = {y | x - 1 / &SUC n < y} INTER {y| y <= x}`
	   by SRW_TAC [] [GSPECIFICATION, EXTENSION]
	   THEN ` {y | x - 1 / &SUC n < y} IN subsets borel` by METIS_TAC [borel_closed, borel_open, OPEN_INTERVAL_RIGHT, OPEN_INTERVAL_LEFT, CLOSED_INTERVAL_RIGHT, CLOSED_INTERVAL_LEFT]
	   THEN `{y | y <= x} IN subsets borel` by METIS_TAC [borel_closed, borel_open, OPEN_INTERVAL_RIGHT, OPEN_INTERVAL_LEFT, CLOSED_INTERVAL_RIGHT, CLOSED_INTERVAL_LEFT]
	   THEN METIS_TAC [sigma_algebra_borel, algebra_def, sigma_algebra_def, ALGEBRA_ALT_INTER])
THEN DISCH_TAC	   
THEN `{y | y <= x} IN subsets borel` by  METIS_TAC [borel_closed, borel_open, OPEN_INTERVAL_RIGHT, OPEN_INTERVAL_LEFT, CLOSED_INTERVAL_RIGHT, CLOSED_INTERVAL_LEFT]
THEN `{y | y < x} IN subsets borel` by  METIS_TAC [borel_closed, borel_open, OPEN_INTERVAL_RIGHT, OPEN_INTERVAL_LEFT, CLOSED_INTERVAL_RIGHT, CLOSED_INTERVAL_LEFT]
THEN `!n. {y | y <= x - 1 / &SUC n} IN subsets borel` by  METIS_TAC [borel_closed, borel_open, OPEN_INTERVAL_RIGHT, OPEN_INTERVAL_LEFT, CLOSED_INTERVAL_RIGHT, CLOSED_INTERVAL_LEFT]
THEN `{y | y = x} IN subsets borel` by  METIS_TAC [borel_closed, borel_open, OPEN_INTERVAL_RIGHT, OPEN_INTERVAL_LEFT, CLOSED_INTERVAL_RIGHT, CLOSED_INTERVAL_LEFT, CLOSED_STANDARD_HYPERPLANE]
THEN METIS_TAC [IN_MEASURABLE, random_variable_def, p_space_def, m_space_def, space_def, events_def, measurable_sets_def, subsets_def]);


val DRBD_event_PREIMAGE = store_thm("DRBD_event_PREIMAGE",
  `` !X p t.
            DRBD_event p X t = PREIMAGE X {s | Normal t < s} INTER p_space p``,
  RW_TAC std_ss[DRBD_event_def]
  THEN SRW_TAC [] [GSPECIFICATION, EXTENSION, PREIMAGE_def]);

val PROB_R_AND = store_thm("PROB_R_AND",
  ``  !p t X Y.
       rv_gt0_ninfinity [X; Y] /\
       indep_var p lborel (\s. real (X s)) lborel (\s. real (Y s))==>
       (prob p (DRBD_event p (R_AND X Y) t) =
        (1-CDF p (\s. real (X s)) t )* (1-CDF p (\s. real (Y s)) t))``,

RW_TAC std_ss [R_AND_INTER, indep_var, DRBD_event_def, rv_gt0_ninfinity_def]
THEN KNOW_TAC(``FINITE {0;1:num }``) THEN1(
     SRW_TAC [] [GSPECIFICATION, EXTENSION, FINITE_DEF])
THEN DISCH_TAC
THEN KNOW_TAC(`` prob p
          (BIGINTER
             {PREIMAGE (if i:num = 0 then (\s. real (X s)) else (\s. real (Y s)))
               {u:real | t < u} INTER p_space p |
             i IN {0:num; 1}}) =
        Normal
          (product {0:num; 1}
             (\i.
                  real
                    (prob p
                       (PREIMAGE
                          (if i:num = 0 then (\s. real (X s))
                           else (\s. real (Y s))) {u:real | t < u} INTER p_space p)))) ``)
      THEN1(
	   DEP_REWRITE_TAC [indep_varsD] 
	   THEN Q.EXISTS_TAC `\i.lborel`
	   THEN Q.EXISTS_TAC `{0;1:num}`
	   THEN RW_TAC std_ss[]
	   THEN1(
		FULL_SIMP_TAC std_ss [indep_vars, indep_sets])
	   THEN1(
		KNOW_TAC(``{0;1:num} = {x|(x=0:num)\/(x=1)}``) THEN1(
		        SRW_TAC [] [GSPECIFICATION, EXTENSION])
		THEN DISCH_TAC
		THEN ASM_REWRITE_TAC[])
	   THEN1(SRW_TAC [] [GSPECIFICATION, EXTENSION])
	   THEN1(SRW_TAC [] [GSPECIFICATION, EXTENSION])
	   THEN RW_TAC std_ss [lborel_gt])
THEN DISCH_TAC
THEN KNOW_TAC(`` {PREIMAGE (if i = 0:num then (\s. real (X s)) else (\s. real (Y s)))
                    {u | t<u} INTER p_space p |
               i IN {0; 1:num}} =
	       {PREIMAGE (\s. real ((X:'a->extreal) s)) {u|t:real<u}  INTER p_space p;
	        PREIMAGE (\s. real ((Y:'a->extreal) s)){u |t < u} INTER p_space p }``)
	      THEN1(SRW_TAC [] [GSPECIFICATION, EXTENSION]
 	            THEN EQ_TAC
		    THEN1( SRW_TAC [] [GSPECIFICATION, EXTENSION]
		    	   THEN1(SRW_TAC [] [GSPECIFICATION, EXTENSION])
			   THEN SRW_TAC [] [GSPECIFICATION, EXTENSION])
	           THEN SRW_TAC [] [GSPECIFICATION, EXTENSION]
			THEN1(Q.EXISTS_TAC `0:num`
			      THEN RW_TAC std_ss[])
			THEN Q.EXISTS_TAC `1:num`
			THEN RW_TAC std_ss[])
THEN DISCH_TAC
THEN FULL_SIMP_TAC std_ss[]
THEN FULL_SIMP_TAC std_ss [BIGINTER_INTER]
THEN KNOW_TAC(``FINITE {0:num} /\ FINITE {1:num} /\ DISJOINT {0:num} {1:num} /\ ({0:num;1} =
                 {0} UNION {1})``)
     THEN1( SRW_TAC [] [GSPECIFICATION, EXTENSION, FINITE_DEF, DISJOINT_DEF])
THEN DISCH_TAC
THEN FULL_SIMP_TAC  std_ss [PRODUCT_UNION]
THEN FULL_SIMP_TAC std_ss  [PRODUCT_SING]
THEN KNOW_TAC(``random_variable (\s. real ((X:'a->extreal) s)) p borel /\
                random_variable  (\s. real (Y s)) p borel``)
     THEN1(FULL_SIMP_TAC std_ss [indep_vars] 
     	   THEN UNDISCH_TAC(`` !i.
        i:num IN {x | (x = 0) \/ (x = 1)} ==>
        random_variable
          (if i = 0 then (\s. real (X s)) else (\s. real (Y s))) p
          (m_space lborel,measurable_sets lborel)``)
	  THEN DISCH_TAC
	  THEN RW_TAC std_ss []
	  THEN1(POP_ASSUM (MP_TAC o Q.SPEC `0:num `)
	  	THEN SRW_TAC [] [GSPECIFICATION, EXTENSION]
		THEN METIS_TAC [subsets_def, measurable_sets_def, borel, lborel, sigma_def, m_space_def, space_def])
	THEN POP_ASSUM (MP_TAC o Q.SPEC `1:num `)
	THEN SRW_TAC [] [GSPECIFICATION, EXTENSION]
		THEN METIS_TAC [subsets_def, measurable_sets_def, borel, lborel, sigma_def, m_space_def, space_def])
THEN DISCH_TAC
THEN KNOW_TAC(``(PREIMAGE (\s. real ((Y:'a->extreal) s)) {u | t:real < u} INTER p_space p)
                  IN events p``)
     THEN1( REWRITE_TAC [events_def]
               THEN KNOW_TAC(``measurable_sets p = subsets (p_space p, measurable_sets p)``) THEN1(
	       	    		RW_TAC std_ss [p_space_def, measurable_sets_def, subsets_def])
		THEN DISCH_TAC THEN ONCE_ASM_REWRITE_TAC[]
		THEN REWRITE_TAC [p_space_def, m_space_def]
		THEN METIS_TAC [random_variable_def, IN_MEASURABLE, events_def, subsets_def, p_space_def, measurable_sets_def, EVENTS_INTER, EVENTS_SPACE, m_space_def, space_def, events_def, lborel_gt,lborel, borel, sigma_def])
THEN KNOW_TAC(``(PREIMAGE (\s. real ((X:'a->extreal) s)) {u | t:real < u} INTER p_space p)
                  IN events p``)
      THEN1( REWRITE_TAC [events_def]
               THEN KNOW_TAC(``measurable_sets p = subsets (p_space p, measurable_sets p)``)
	            THEN1(
	       	    		RW_TAC std_ss [p_space_def, measurable_sets_def, subsets_def])
		THEN DISCH_TAC THEN ONCE_ASM_REWRITE_TAC[]
		THEN REWRITE_TAC [p_space_def, m_space_def]
		THEN METIS_TAC [random_variable_def, IN_MEASURABLE, events_def, subsets_def, p_space_def, measurable_sets_def, EVENTS_INTER, EVENTS_SPACE, m_space_def, space_def, events_def, lborel_gt,lborel, borel, sigma_def])
THEN DISCH_TAC
THEN DISCH_TAC

THEN KNOW_TAC(``prob_space p``) THEN1(FULL_SIMP_TAC std_ss [indep_vars, indep_sets])
THEN DISCH_TAC
 THEN FULL_SIMP_TAC std_ss [normal_real_mul, PROB_FINITE]
THEN FULL_SIMP_TAC std_ss [normal_real, PROB_FINITE]
THEN KNOW_TAC(``{s | Normal t < (X:'a->extreal) s} = PREIMAGE X {s| Normal t <  s}``) THEN1(SRW_TAC [] [GSPECIFICATION, EXTENSION, PREIMAGE_def])
THEN DISCH_TAC
THEN KNOW_TAC(``{s | Normal t < (Y:'a->extreal) s} = PREIMAGE Y {s| Normal t < s}``) THEN1(SRW_TAC [] [GSPECIFICATION, EXTENSION, PREIMAGE_def])
THEN DISCH_TAC
THEN ASM_REWRITE_TAC[]
THEN RW_TAC std_ss[PREIMAGE_EXTREAL_REAL_lt]
THEN RW_TAC std_ss[GSYM PREIMAGE_EXTREAL_REAL_lt]
THEN RW_TAC std_ss[GSYM DRBD_event_PREIMAGE]
THEN RW_TAC std_ss[PROB_DRBD_DFT_COMP, rv_gt0_ninfinity_def]
THEN RW_TAC std_ss[CDF_prob_DFT_event]);


val Rel_R_AND = store_thm("Rel_R_AND",
  ``  !p t X Y.
       rv_gt0_ninfinity [X; Y] /\
       indep_var p lborel (real o X) lborel (real o Y)==>
       (Rel p (R_AND X Y) t =
        Rel p X t * Rel p Y t)``,

RW_TAC std_ss[Rel_def, o_DEF]
THEN DEP_REWRITE_TAC[PROB_R_AND]
THEN RW_TAC std_ss[]
THEN DEP_REWRITE_TAC[PROB_DRBD_DFT_COMP]
THEN RW_TAC std_ss[]
     THEN1(FULL_SIMP_TAC std_ss[rv_gt0_ninfinity_def])
     THEN1(`borel = (m_space lborel, measurable_sets lborel)` by METIS_TAC[lborel, borel, m_space_def, space_def, measurable_sets_def, subsets_def, PAIR, space_borel]
          THEN METIS_TAC[indep_var_rv])
     THEN1(FULL_SIMP_TAC std_ss[rv_gt0_ninfinity_def])
     THEN1(`borel = (m_space lborel, measurable_sets lborel)` by METIS_TAC[lborel, borel, m_space_def, space_def, measurable_sets_def, subsets_def, PAIR, space_borel]
          THEN METIS_TAC[indep_var_rv])

THEN `borel = (m_space lborel, measurable_sets lborel)` by METIS_TAC[lborel, borel, m_space_def, space_def, measurable_sets_def, subsets_def, PAIR, space_borel]
THEN `rv_gt0_ninfinity[X]` by FULL_SIMP_TAC std_ss[rv_gt0_ninfinity_def]
THEN `rv_gt0_ninfinity[X]` by FULL_SIMP_TAC std_ss[rv_gt0_ninfinity_def]
THEN `random_variable (\s. real (X s)) p borel` by METIS_TAC[indep_var_rv]
THEN `random_variable (\s. real (Y s)) p borel` by METIS_TAC[indep_var_rv]
THEN METIS_TAC[CDF_prob_DFT_event, rv_gt0_ninfinity_def,indep_var_rv]);



val PROB_R_OR = store_thm("PROB_R_OR",
  ``  !p t X Y.
       rv_gt0_ninfinity [X; Y] /\
       indep_var p lborel (\s. real (X s)) lborel (\s. real (Y s))   ==>
       (prob p (DRBD_event p (R_OR X Y) t) =
        (1-CDF p (\s. real (X s)) t * CDF p (\s. real (Y s)) t))``,

RW_TAC std_ss[R_OR_UNION]
THEN KNOW_TAC(`` DRBD_event p X t UNION DRBD_event p Y t =
                 p_space p DIFF (p_space p DIFF (DRBD_event p X t UNION DRBD_event p Y t))``)

    THEN1(SRW_TAC[] [GSPECIFICATION, EXTENSION]
          THEN FULL_SIMP_TAC std_ss[DRBD_event_def] THEN ASM_SET_TAC[])
THEN RW_TAC std_ss[]
THEN ONCE_ASM_REWRITE_TAC[]
THEN KNOW_TAC(``prob p
     (p_space p DIFF
      (p_space p DIFF (DRBD_event p X t UNION DRBD_event p Y t))) =
      1- prob p
           (p_space p DIFF (DRBD_event p X t UNION DRBD_event p Y t))``)
    THEN1(DEP_REWRITE_TAC[PROB_COMPL]
          THEN FULL_SIMP_TAC std_ss[indep_var, indep_sets, indep_vars]
	  THEN KNOW_TAC(``random_variable (\s. real (X s)) p borel ``)
	       THEN1(UNDISCH_TAC(``
	      !i:num.
               i IN {x | (x = 0) \/ (x = 1)} ==>
               random_variable
               (if i = 0 then (\s. real (X s)) else (\s. real (Y s))) p
               (m_space lborel,measurable_sets lborel) ``)
	            THEN DISCH_TAC
		    THEN POP_ASSUM (MP_TAC o Q.SPEC `0:num `)
                    THEN PSET_TAC[]
		    THEN `borel = (m_space lborel, measurable_sets lborel)` by METIS_TAC[borel, lborel, m_space_def, measurable_sets_def, PAIR, space_def, subsets_def, space_borel]
		    THEN FULL_SIMP_TAC std_ss[])
          THEN DISCH_TAC
	  THEN  KNOW_TAC(``random_variable (\s. real (Y s)) p borel ``)
	       THEN1(UNDISCH_TAC(``
	       !i:num.
               i IN {x | (x = 0) \/ (x = 1)} ==>
               random_variable
               (if i = 0 then (\s. real (X s)) else (\s. real (Y s))) p
               (m_space lborel,measurable_sets lborel) ``)
	            THEN DISCH_TAC
		    THEN POP_ASSUM (MP_TAC o Q.SPEC `1:num `)
                    THEN PSET_TAC[]
		    THEN `borel = (m_space lborel, measurable_sets lborel)` by METIS_TAC[borel, lborel, m_space_def, measurable_sets_def, PAIR, space_def, subsets_def, space_borel]
		    THEN FULL_SIMP_TAC std_ss[])
          THEN DISCH_TAC
          THEN METIS_TAC[rv_gt0_ninfinity_def,EVENTS_COMPL, EVENTS_UNION, DRBD_event_in_event_p])
THEN DISCH_TAC
THEN ONCE_ASM_REWRITE_TAC[]
THEN KNOW_TAC(`` p_space p DIFF (DRBD_event p X t UNION DRBD_event p Y t) =
                (p_space p DIFF DRBD_event p X t) INTER(p_space p DIFF DRBD_event p Y t)``)

    THEN1(SRW_TAC[][GSPECIFICATION, EXTENSION]
          THEN RW_TAC std_ss[]
	  THEN EQ_TAC THEN RW_TAC std_ss[])
THEN DISCH_TAC THEN ONCE_ASM_REWRITE_TAC[]	  
THEN KNOW_TAC(``  prob p
     ((p_space p DIFF DRBD_event p X t) INTER
      (p_space p DIFF DRBD_event p Y t)) =
      prob p (p_space p DIFF DRBD_event p X t) *
      prob p (p_space p DIFF DRBD_event p Y t)``)
    THEN1(RW_TAC std_ss[DRBD_event_def]
          THEN `{s | Normal t < X s} INTER p_space p =
       PREIMAGE X {s | Normal t < s} INTER p_space p` by ASM_SET_TAC[PREIMAGE_def]
          THEN `{s | Normal t < Y s} INTER p_space p =
       PREIMAGE Y {s | Normal t < s} INTER p_space p` by ASM_SET_TAC[PREIMAGE_def]
       THEN ONCE_ASM_REWRITE_TAC[]
       THEN KNOW_TAC(`` p_space p DIFF PREIMAGE X {s | Normal t < s} INTER p_space p=
                        PREIMAGE X {s | s<= Normal t} INTER p_space p``)
            THEN1(SRW_TAC[][GSPECIFICATION, EXTENSION, PREIMAGE_def]
	          THEN EQ_TAC
		       THEN1(RW_TAC std_ss[]
		             THEN METIS_TAC[extreal_lt_def, extreal_le_def, lt_le, le_lt, lt_total, le_total])
	          THEN RW_TAC std_ss[] THEN METIS_TAC[extreal_lt_def, extreal_le_def, lt_le, le_lt, lt_total, le_total])
      THEN DISCH_TAC
      THEN  KNOW_TAC(`` p_space p DIFF PREIMAGE Y {s | Normal t < s} INTER p_space p=
                        PREIMAGE Y {s | s<= Normal t} INTER p_space p``)
            THEN1(SRW_TAC[][GSPECIFICATION, EXTENSION, PREIMAGE_def]
	          THEN EQ_TAC
		       THEN1(RW_TAC std_ss[]
		             THEN METIS_TAC[extreal_lt_def, extreal_le_def, lt_le, le_lt, lt_total, le_total])
	          THEN RW_TAC std_ss[] THEN METIS_TAC[extreal_lt_def, extreal_le_def, lt_le, le_lt, lt_total, le_total])
      THEN DISCH_TAC THEN ASM_REWRITE_TAC[]
      THEN MP_TAC(Q.ISPECL [`p:('a -> bool) # (('a -> bool) -> bool) # (('a -> bool) -> extreal)`,`(\i:num. lborel):num->((real->bool)#((real->bool)->bool)#((real->bool)->extreal))`, `(\i:num. if i = 0 then (\s. real ((X:'a->extreal) s)) else (\s. real (Y s)))`,`{0:num; 1}`,` (\i. if i = 0:num then {s | s:real <=  t} else {s | s <=  t})`, `0:num`, `1:num`]  prob_prod_any_2_of_10)
      THEN RW_TAC std_ss[]
      THEN `prob_space p` by FULL_SIMP_TAC std_ss[indep_var, indep_vars, indep_sets]
      THEN `{x | (x = 0) \/ (x = 1:num)} = {0;1:num}` by ASM_SET_TAC[]
      THEN FULL_SIMP_TAC std_ss[]
      THEN `{0:num; 1} SUBSET {0; 1}` by ASM_SET_TAC[]
      THEN  KNOW_TAC(`` (!i:num. i IN {0; 1} ==>
                        {s | s <= t} IN measurable_sets lborel)``)
            THEN1(RW_TAC std_ss[]
	          THEN `measurable_sets lborel = subsets borel` by METIS_TAC[borel, lborel, m_space_def, space_def,measurable_sets_def, subsets_def, space_borel]
		  THEN RW_TAC std_ss[] THEN METIS_TAC[borel_closed, CLOSED_INTERVAL_RIGHT, CLOSED_INTERVAL_LEFT])
		  
     THEN DISCH_TAC THEN FULL_SIMP_TAC std_ss[]
     THEN `{s | s <= t} IN measurable_sets lborel` by ( `measurable_sets lborel = subsets borel` by METIS_TAC[borel, lborel, m_space_def, space_def,measurable_sets_def, subsets_def, space_borel]
		  THEN RW_TAC std_ss[] THEN METIS_TAC[borel_closed, CLOSED_INTERVAL_RIGHT, CLOSED_INTERVAL_LEFT])
     THEN FULL_SIMP_TAC std_ss[indep_var]
     THEN `PREIMAGE X {s | s <= Normal t} INTER p_space p =
           PREIMAGE (\s. real (X s)) {s | s <= t} INTER p_space p` by METIS_TAC[PREIMAGE_EXTREAL_REAL, rv_gt0_ninfinity_def]
     THEN `PREIMAGE Y {s | s <= Normal t} INTER p_space p =
           PREIMAGE (\s. real (Y s)) {s | s <= t} INTER p_space p` by METIS_TAC[PREIMAGE_EXTREAL_REAL, rv_gt0_ninfinity_def]
     THEN RW_TAC std_ss[])
THEN RW_TAC std_ss[]
THEN RW_TAC std_ss[DRBD_event_def]
THEN KNOW_TAC(`` p_space p DIFF PREIMAGE X {s | Normal t < s} INTER p_space p =
                 PREIMAGE X {s | s<= Normal t} INTER p_space p``)
     THEN1(SRW_TAC[][GSPECIFICATION, EXTENSION, PREIMAGE_def]
           THEN EQ_TAC
 	       THEN1(RW_TAC std_ss[]
		     THEN METIS_TAC[extreal_lt_def, extreal_le_def, lt_le, le_lt, lt_total, le_total])
	   THEN RW_TAC std_ss[] THEN METIS_TAC[extreal_lt_def, extreal_le_def, lt_le, le_lt, lt_total, le_total])
THEN DISCH_TAC
THEN  KNOW_TAC(`` p_space p DIFF PREIMAGE Y {s | Normal t < s} INTER p_space p =
                  PREIMAGE Y {s | s<= Normal t} INTER p_space p``)
      THEN1(SRW_TAC[][GSPECIFICATION, EXTENSION, PREIMAGE_def]
           THEN EQ_TAC
               THEN1(RW_TAC std_ss[]
	            THEN METIS_TAC[extreal_lt_def, extreal_le_def, lt_le, le_lt, lt_total, le_total])
	  THEN RW_TAC std_ss[] THEN METIS_TAC[extreal_lt_def, extreal_le_def, lt_le, le_lt, lt_total, le_total])
THEN RW_TAC std_ss[]
THEN ` {s | Normal t < X s} INTER p_space p =
       PREIMAGE X {s | Normal t < s} INTER p_space p` by ASM_SET_TAC[PREIMAGE_def]
THEN ` {s | Normal t < Y s} INTER p_space p =
       PREIMAGE Y {s | Normal t < s} INTER p_space p` by ASM_SET_TAC[PREIMAGE_def]
THEN ASM_REWRITE_TAC[]
THEN `PREIMAGE X {s | s <= Normal t} INTER p_space p =
      PREIMAGE (\s. real (X s)) {s | s <=  t} INTER p_space p` by METIS_TAC[PREIMAGE_EXTREAL_REAL, rv_gt0_ninfinity_def]
THEN  `PREIMAGE Y {s | s <= Normal t} INTER p_space p =
       PREIMAGE (\s. real (Y s)) {s | s <=  t} INTER p_space p` by METIS_TAC[PREIMAGE_EXTREAL_REAL, rv_gt0_ninfinity_def]

THEN FULL_SIMP_TAC std_ss[indep_var, indep_vars]
THEN `PREIMAGE X {s | s <= Normal t} INTER p_space p =
      DFT_event p X t` by ASM_SET_TAC[DFT_event_def, PREIMAGE_def]
THEN `PREIMAGE Y {s | s <= Normal t} INTER p_space p =
      DFT_event p Y t` by ASM_SET_TAC[DFT_event_def, PREIMAGE_def]

THEN ONCE_ASM_REWRITE_TAC[]
THEN METIS_TAC[CDF_prob_DFT_event, rv_gt0_ninfinity_def]);


val DFT_event_in_event_p = store_thm("DFT_event_in_event_p",
  `` !X p t. rv_gt0_ninfinity[X] /\  random_variable (\s. real (X s)) p borel ==>
             DFT_event p X t IN events p``,

RW_TAC std_ss[DFT_event_def, rv_gt0_ninfinity_def]
THEN ` {s | X s <= Normal t} INTER p_space p =
       PREIMAGE X {s | s <= Normal t} INTER p_space p` by ASM_SET_TAC[PREIMAGE_def]
THEN RW_TAC std_ss[PREIMAGE_EXTREAL_REAL]
THEN METIS_TAC[random_variable_def,borel_closed, CLOSED_INTERVAL_LEFT, p_space_def, m_space_def, space_def, events_def, measurable_sets_def, subsets_def, IN_MEASURABLE]);


val Rel_R_OR = store_thm("Rel_R_OR",
  ``   !p t X Y.
          rv_gt0_ninfinity [X; Y] /\
          indep_var p lborel (real o X) lborel (real o Y) ==>
          (Rel p (R_OR X Y) t =
           1 - (1 - Rel p X t) * (1 - Rel p Y t))``,


RW_TAC std_ss[Rel_def, o_DEF]
THEN DEP_REWRITE_TAC[PROB_R_OR]
THEN RW_TAC std_ss[]
THEN KNOW_TAC(``1 - prob p (DRBD_event p X t) = prob p (DFT_event p X t) ``)
    THEN1(DEP_REWRITE_TAC[PROB_DFT_DRBD_COMP]
          THEN FULL_SIMP_TAC std_ss[rv_gt0_ninfinity_def]
	  THEN `borel = (m_space lborel, measurable_sets lborel)` by METIS_TAC[lborel, borel, m_space_def, space_def, measurable_sets_def, subsets_def, PAIR, space_borel]
          THEN METIS_TAC[indep_var_rv])
THEN DISCH_TAC THEN ONCE_ASM_REWRITE_TAC[]
THEN KNOW_TAC(``1 - prob p (DRBD_event p Y t) = prob p (DFT_event p Y t) ``)
    THEN1(DEP_REWRITE_TAC[PROB_DFT_DRBD_COMP]
          THEN FULL_SIMP_TAC std_ss[rv_gt0_ninfinity_def]
	  THEN `borel = (m_space lborel, measurable_sets lborel)` by METIS_TAC[lborel, borel, m_space_def, space_def, measurable_sets_def, subsets_def, PAIR, space_borel]
          THEN METIS_TAC[indep_var_rv])
THEN DISCH_TAC THEN ONCE_ASM_REWRITE_TAC[]
THEN METIS_TAC[CDF_prob_DFT_event, rv_gt0_ninfinity_def]);




(******************************************************************************)
(*                 required properties of indep_sets                          *)
(*                                                                            *)
(*                                                                            *)
(******************************************************************************)

val indep_sets_collect_sigma_BIGUNION = store_thm(
  "indep_sets_collect_sigma_BIGUNION",
  `` !p E J s.
            indep_sets p
              (\j. sigma_sets (m_space p) (BIGUNION {E i | i IN s j})) J ==>
            indep_sets p (\j. BIGUNION {E i | i IN s j}) J``,

RW_TAC std_ss[]
THEN MATCH_MP_TAC  indep_sets_mono_sets
THEN Q.EXISTS_TAC `(\j. sigma_sets (m_space p) (BIGUNION {E i | i IN s j}))`
THEN RW_TAC std_ss[]
THEN RW_TAC  std_ss[sigma_sets_superset_generator]);


val indep_sets_sigma_compl = store_thm(
  "indep_sets_sigma_compl",
  `` !p E J.
            indep_sets p (\j. sigma_sets (m_space p) {E j}) J ==>
            indep_sets p (\j. {m_space p DIFF E j}) J``,

RW_TAC std_ss[]
THEN MATCH_MP_TAC  indep_sets_mono_sets
THEN Q.EXISTS_TAC `(\j. sigma_sets (m_space p) {E j})`
THEN RW_TAC std_ss[]
THEN KNOW_TAC(`` E (i:num) IN sigma_sets (m_space p) {E i }``)
     THEN1(`E (i:num) IN {E i}` by ASM_SET_TAC[]
           THEN METIS_TAC[sigma_sets_basic])
THEN RW_TAC std_ss[]	   
THEN `{E (i:num)} SUBSET sigma_sets (m_space p) {E i}` by ASM_SET_TAC[]
THEN KNOW_TAC(``m_space p DIFF E (i:num) IN sigma_sets (m_space p) {E i}``)
      THEN1(MATCH_MP_TAC sigma_sets_compl
            THEN RW_TAC std_ss[])
THEN ASM_SET_TAC[] );


val BIGINTER_SUBSET_BIGUNION = store_thm("BIGINTER_SUBSET_BIGUNION",
  ``!P. P <> {} ==>  BIGINTER P  SUBSET BIGUNION P ``, 
ASM_SET_TAC[]);

val indep_sets_BIGUNION_BIGINTER = store_thm("indep_sets_BIGUNION_BIGINTER",
  ``  !E J s p.
            (!i. s i <> {}) /\
            indep_sets p (\j. BIGUNION {E i | i IN s j}) J ==>
            indep_sets p (\j. BIGINTER {E i | i IN s j}) J``,

RW_TAC std_ss[]
THEN MATCH_MP_TAC  indep_sets_mono_sets
THEN Q.EXISTS_TAC `(\j. BIGUNION {E i | i IN s j})`
THEN RW_TAC std_ss[]
THEN `{E i' | i' IN s i} <> {}` by ASM_SET_TAC[]
THEN RW_TAC std_ss[]
THEN ASM_SET_TAC[BIGINTER_SUBSET_BIGUNION]);



(* ========================================================================== *)
(*                  Probability of series and parallel structures             *)
(*                                                                            *)
(*                                                                            *)
(* ========================================================================== *)

val DRBD_series_def = Define `(DRBD_series  Y (s:num->bool) = BIGINTER {Y i | i IN s}  ) `;

val DRBD_parallel_def = Define `(DRBD_parallel Y (s:num->bool) = BIGUNION {Y i | i IN s}) `;


val DRBD_series_AND_INTER = store_thm("DRBD_series_AND_INTER",
  ``  !X Y p t.
            DRBD_event p (R_AND X Y) t =
            DRBD_series
              (\i. if i = 0 then DRBD_event p X t else DRBD_event p Y t)
              {0; 1}  ``,

RW_TAC std_ss[R_AND_INTER]
THEN RW_TAC std_ss[DRBD_series_def]
THEN KNOW_TAC(`` BIGINTER
     {if i:num = 0 then DRBD_event p X t else DRBD_event p Y t | i IN {0; 1}} =
     BIGINTER
     {DRBD_event p X t ; DRBD_event p Y t }``)

     THEN1(AP_TERM_TAC
           THEN SRW_TAC[][GSPECIFICATION, EXTENSION]
	   THEN EQ_TAC
	   THEN RW_TAC std_ss[]
	        THEN1 ASM_SET_TAC[]
		THEN1 ASM_SET_TAC[]
		THEN1 ASM_SET_TAC[]
		THEN Q.EXISTS_TAC`1` THEN RW_TAC std_ss[])
THEN RW_TAC std_ss[]
THEN METIS_TAC[BIGINTER_INTER]);


val DRBD_series_AND_INTER_3  = store_thm("DRBD_series_AND_INRER_3",
  ``  !X Y Z p t.
            DRBD_event p (R_AND (R_AND X Y) Z) t =
            DRBD_series
              (\i.
                   if i = 0 then DRBD_event p X t
                   else if i = 1 then DRBD_event p Y t
                   else DRBD_event p Z t) {0; 1; 2} ``,

RW_TAC std_ss[R_AND_INTER]
THEN RW_TAC std_ss[DRBD_series_def]
THEN KNOW_TAC(`` BIGINTER
     {if i = 0 then DRBD_event p X t
     else if i = 1 then DRBD_event p Y t
     else DRBD_event p Z t |
     i IN {0; 1; 2:num}} =
     BIGINTER
     {DRBD_event p X t ; DRBD_event p Y t; DRBD_event p Z t }``)

     THEN1(AP_TERM_TAC
           THEN SRW_TAC[][GSPECIFICATION, EXTENSION]
	   THEN EQ_TAC
	   THEN RW_TAC std_ss[]
	        THEN1 ASM_SET_TAC[]
		THEN1 ASM_SET_TAC[]
		THEN1 ASM_SET_TAC[]
		THEN1 ASM_SET_TAC[]
	        THEN1 (Q.EXISTS_TAC`1` THEN RW_TAC std_ss[])
 	   THEN Q.EXISTS_TAC`2` THEN RW_TAC std_ss[])

THEN RW_TAC std_ss[]
THEN METIS_TAC[BIGINTER_3_sets]);


val DRBD_event_INTER_p_space_p = store_thm("DRBD_event_INTER_p_space_p",
  ``!p X t. DRBD_event p X t INTER p_space p = DRBD_event p X t  ``,

RW_TAC std_ss[DRBD_event_def]
THEN ASM_SET_TAC[]);

val rv_to_event_def = Define `rv_to_event p X t = (\i. DRBD_event p (X i) t )  `;


val PROB_DRBD_series = store_thm("PROB_DRBD_series",
  `` 
        !p X s.
            indep_sets p (\i. {X i}) s /\ s <> {} /\ FINITE s ==>
            (prob p (DRBD_series  X s) =
             Normal (product s (\i. real (prob p (X i)))))``,

RW_TAC std_ss[DRBD_series_def]
THEN MATCH_MP_TAC indep_setsD
THEN Q.EXISTS_TAC `\i. {X i}`
THEN Q.EXISTS_TAC`s`
THEN RW_TAC std_ss[]
     THEN1(FULL_SIMP_TAC std_ss[indep_sets])
THEN ASM_SET_TAC[]);


val PROB_DRBD_series_rv = store_thm("PROB_DRBD_series_rv",
  `` !p X t s.
            s <> {} /\ FINITE s /\
            indep_sets p (\i. {rv_to_event p X t i}) s  ==>
            (prob p (DRBD_series (rv_to_event p X t) s) =
             Normal (product s (\i. real (Rel p (X i) t))))``,

RW_TAC std_ss[rv_to_event_def]
THEN DEP_REWRITE_TAC[PROB_DRBD_series]
THEN RW_TAC std_ss[Rel_def]);



val BIGUNION_IN_events_p = store_thm("BIGUNION_IN_events_p",
  ``!p X s.FINITE (s:num->bool) /\ (!i. i IN s ==> X i IN events p) /\ prob_space p ==>
             BIGUNION{X i| i IN s} IN events p``,
RW_TAC std_ss[]
THEN MATCH_MP_TAC EVENTS_COUNTABLE_UNION
THEN RW_TAC std_ss[]
     THEN1(ASM_SET_TAC[])
THEN MATCH_MP_TAC finite_countable
THEN RW_TAC std_ss[GSYM IMAGE_DEF]
THEN RW_TAC std_ss[IMAGE_FINITE]);




val PROB_DRBD_parallel_compl = store_thm("PROB_DRBD_parallel_compl",
  ``  !p X s.
            indep_sets p (\i. {X i}) s /\ FINITE s ==>
            (prob p (DRBD_parallel X s) =
             1 - prob p (p_space p DIFF BIGUNION {X i | i IN s}))``,

RW_TAC std_ss[DRBD_parallel_def]
THEN KNOW_TAC(``(!i:num. i IN s ==> X i IN events p)      ``)
     THEN1(FULL_SIMP_TAC std_ss[indep_sets]
           THEN ASM_SET_TAC[])
THEN RW_TAC std_ss[]	   
THEN DEP_REWRITE_TAC[GSYM PROB_COMPL]
THEN RW_TAC std_ss[]
     THEN1(FULL_SIMP_TAC std_ss[indep_sets])
     THEN1(`prob_space p`  by FULL_SIMP_TAC std_ss[indep_sets]
           THEN `BIGUNION {X i | i:num IN s} IN events p` by RW_TAC std_ss[BIGUNION_IN_events_p]
	   THEN RW_TAC std_ss[EVENTS_COMPL])
THEN KNOW_TAC(``BIGUNION {X i | i:num IN s} =
                p_space p DIFF (p_space p DIFF BIGUNION {X i | i IN s})``) 	   
     THEN1(KNOW_TAC(`` BIGUNION {X i | i:num IN s} SUBSET p_space p``)
           THEN1(RW_TAC std_ss[p_space_def]
	        THEN `measure_space p` by FULL_SIMP_TAC std_ss[indep_sets, prob_space_def]
		THEN `BIGUNION{X i| i IN s} IN measurable_sets p` by METIS_TAC[BIGUNION_IN_events_p, events_def, indep_sets]
		THEN RW_TAC std_ss[MEASURABLE_SETS_SUBSET_SPACE])
	   THEN RW_TAC std_ss[DIFF_DIFF_SUBSET])
THEN RW_TAC std_ss[]
THEN METIS_TAC[]);



 val DIFF_BIGINTER2 = store_thm( (* from util_prob *)
     "DIFF_BIGINTER2",
      ``!sp s.
            (!t. t IN s ==> t SUBSET sp) /\ s <> {} ==>
            (BIGINTER s = sp DIFF (BIGUNION (IMAGE (\u. sp DIFF u) s)))``,
RW_TAC std_ss []
THEN `(BIGINTER s SUBSET sp)` by RW_TAC std_ss [BIGINTER_SUBSET]
THEN ASSUME_TAC (Q.SPECL [`sp`,`s`] DIFF_BIGINTER1)
THEN  `sp DIFF (sp DIFF (BIGINTER s)) = (BIGINTER s)` by RW_TAC std_ss [DIFF_DIFF_SUBSET]
THEN  METIS_TAC []);

val DIFF_BIGUNION_BIGINTER = store_thm("DIFF_BIGUNION_BIGINTER",
  `` !p X s.
            s <> {} /\ (!i:num. i IN s ==> X i IN events p) /\ prob_space p ==>
            (p_space p DIFF BIGUNION {X i | i IN s} =
             BIGINTER {p_space p DIFF X i | i IN s})``,

RW_TAC std_ss[]
THEN KNOW_TAC(``!t. t IN {X i | i:num IN s}==> t SUBSET p_space p ``)
     THEN1(KNOW_TAC(``!a. a IN events p ==> a SUBSET p_space p ``)
           THEN1(RW_TAC std_ss[p_space_def]
	   	 THEN FULL_SIMP_TAC std_ss[events_def, prob_space_def]
		 THEN RW_TAC std_ss[MEASURABLE_SETS_SUBSET_SPACE])
	   THEN RW_TAC std_ss[]	 
	   THEN KNOW_TAC(`` t IN events p``)
	   	THEN1(ASM_SET_TAC[])
	   THEN RW_TAC std_ss[])

THEN RW_TAC std_ss[]
THEN `{p_space p DIFF X i| i:num IN s} <> {}` by ASM_SET_TAC[]
THEN KNOW_TAC(``!t. t IN {p_space p DIFF X i | i:num IN s}==> t SUBSET p_space p ``)
     THEN1(RW_TAC std_ss[]
          THEN KNOW_TAC(``t IN events p ``)
	       THEN1(KNOW_TAC(``!i:num. i IN s ==> p_space p DIFF X i IN events p``)
	             THEN1(RW_TAC std_ss[]
		          THEN UNDISCH_TAC(``!i:num. i IN s ==> X i IN events p ``)
			  THEN DISCH_TAC
			  THEN POP_ASSUM (MP_TAC o Q.SPEC `i:num`)
			  THEN RW_TAC std_ss[]
			  THEN METIS_TAC[EVENTS_COMPL])
		     THEN RW_TAC std_ss[]
		     THEN ASM_SET_TAC[])
	  THEN RW_TAC std_ss[p_space_def]
	  THEN FULL_SIMP_TAC std_ss[events_def, prob_space_def]
	  THEN RW_TAC std_ss[MEASURABLE_SETS_SUBSET_SPACE])
THEN RW_TAC std_ss[]
THEN KNOW_TAC(``  BIGUNION {(X:num->'a->bool) (i:num) | i IN s} =
                  BIGUNION {p_space p DIFF (p_space p DIFF X i) | i IN s}``)
     THEN1(KNOW_TAC(``{(X:num->'a->bool) (i:num)|i IN s} =
                      {p_space p DIFF (p_space p DIFF X i)| i IN s}``)
           THEN1(SRW_TAC[][GSPECIFICATION, EXTENSION]
	         THEN EQ_TAC
		       THEN1( RW_TAC std_ss[IN_DIFF]
		             THEN ASM_SET_TAC[IN_DIFF])
		 THEN RW_TAC std_ss[]
		 THEN ASM_SET_TAC[IN_DIFF])
     THEN RW_TAC std_ss[])
THEN RW_TAC std_ss[]     
THEN KNOW_TAC(``{p_space p DIFF (p_space p DIFF X i) | i:num IN s} =
                IMAGE (\u. p_space p DIFF u) {p_space p DIFF X i| i IN s} ``)
    THEN1(ASM_SET_TAC[IMAGE_DEF])
THEN RW_TAC std_ss[]    
THEN  METIS_TAC[DIFF_BIGINTER2]);



val PROB_DRBD_parallel_BIGINTER = store_thm("PROB_DRBD_parallel_BIGINTER",
  ``   !p X s.
            indep_sets p (\i. {X i}) s /\ FINITE s /\ s <> {} ==>
            (prob p (DRBD_parallel X s) =
             1 - prob p (BIGINTER {p_space p DIFF X i | i IN s}))``,

RW_TAC std_ss[]
THEN KNOW_TAC(`` (!i:num. i IN s ==> X i IN events p)``)
     THEN1(FULL_SIMP_TAC std_ss[indep_sets]
          THEN ASM_SET_TAC[])
THEN RW_TAC std_ss[]
THEN DEP_REWRITE_TAC[ PROB_DRBD_parallel_compl]
THEN RW_TAC std_ss[]
THEN `prob_space p` by FULL_SIMP_TAC std_ss[indep_sets]
THEN KNOW_TAC(``p_space p DIFF BIGUNION {X i| i:num IN s} =
                BIGINTER {p_space p DIFF X i| i IN s}``)
    THEN1(DEP_REWRITE_TAC[DIFF_BIGUNION_BIGINTER]
          THEN RW_TAC std_ss[])
THEN RW_TAC std_ss[]);	  


val PROB_BIGINTER_COMPL = store_thm("PROB_BIGINTER_COMPL",
  ``!p X s.
            indep_sets p (\i. {X i}) s /\ FINITE s /\ s <> {} ==>
            (prob p (BIGINTER {p_space p DIFF X i | i IN s}) =
             Normal (product s (\i. real (1 - prob p (X i)))))``,

RW_TAC std_ss[]
THEN KNOW_TAC(``(!i:num. i IN s ==> X i IN events p)``)
     THEN1(FULL_SIMP_TAC std_ss[indep_sets]
          THEN ASM_SET_TAC[])
THEN RW_TAC std_ss[]          
THEN KNOW_TAC(`` indep_sets p (\j. sigma_sets (m_space p) {X (j:num)}) s``)
    THEN1(ASSUME_TAC (Q.SPECL [ `p:('a->bool)#(('a->bool)->bool)#(('a->bool)->extreal)`, `(\i. {X i})`, `s:num->bool`] indep_sets_sigma')
          THEN RW_TAC std_ss[GSYM p_space_def]
          THEN  `prob_space p` by FULL_SIMP_TAC std_ss[indep_sets]
          THEN FULL_SIMP_TAC std_ss[]
          THEN KNOW_TAC(``(!i . i:num IN s ==>  Int_stable {X i})``)
               THEN1(RW_TAC std_ss[]
 	             THEN RW_TAC std_ss[Int_stable]
 		     THEN PSET_TAC[])

          THEN RW_TAC std_ss[]
	  THEN FULL_SIMP_TAC std_ss[])
THEN RW_TAC std_ss[]
THEN DEP_REWRITE_TAC[indep_setsD]
THEN RW_TAC std_ss[]
     THEN1(Q.EXISTS_TAC `(\i. {p_space p DIFF X i}) `
           THEN Q.EXISTS_TAC `s`
           THEN RW_TAC std_ss[]
	        THEN1(FULL_SIMP_TAC std_ss[indep_sets])
		THEN1(RW_TAC std_ss[indep_sets_sigma_compl, p_space_def])
		THEN1( ASM_SET_TAC[])
           THEN ASM_SET_TAC[])
THEN KNOW_TAC(``!i:num. i IN s ==> (prob p (p_space p DIFF X i) = 1-prob p (X i)) ``)	   
    THEN1(RW_TAC std_ss[]
          THEN `X (i:num) IN events p` by RW_TAC std_ss[]
	  THEN `prob_space p ` by FULL_SIMP_TAC std_ss[indep_sets]
	  THEN RW_TAC std_ss[PROB_COMPL])
THEN RW_TAC std_ss[]
THEN `!i:num. i IN s ==>
              (real (prob p (p_space p DIFF X i)) = real (1-prob p (X i)))` by RW_TAC std_ss[]
THEN RW_TAC std_ss[th]);


val PROB_DRBD_parallel = store_thm("PROB_DRBD_parallel",
  ``  !p X s.
            indep_sets p (\i. {X i}) s /\ s <> {} /\ FINITE s ==>
            (prob p (DRBD_parallel  X s) =
             1 - Normal (product s (\i. real (1 - prob p (X i)))))``,
RW_TAC std_ss[PROB_DRBD_parallel_BIGINTER]
THEN METIS_TAC[PROB_BIGINTER_COMPL]);



val PROB_DRBD_parallel_rv = store_thm("PROB_DRBD_parallel_rv",
  ``  !p X t s.
            s <> {} /\ FINITE s /\
            indep_sets p (\i. {rv_to_event p X t i}) s ==>
            (prob p (DRBD_parallel  (rv_to_event p X t) s) =
             1 - Normal (product s (\i. real (1 - Rel p (X i) t))))``,

RW_TAC std_ss[rv_to_event_def, Rel_def]
THEN DEP_REWRITE_TAC[PROB_DRBD_parallel]
THEN RW_TAC std_ss[]);

  

(* ========================================================================== *)
(*                                                                            *)
(*                       n-ary AND and n-ary OR                               *)
(*                                                                            *)
(* ========================================================================== *)

val nR_AND_def = Define `nR_AND X s = pred_set$ITSET (\e acc. R_AND(X e) acc) s R_NEVER `;

val nR_AND_THM = store_thm("nR_AND_THM",
  ``!X. (nR_AND X {} = (\s. PosInf) )/\
        (!e s. FINITE s ==>
	        (nR_AND X (e INSERT s) = R_AND (X e) (nR_AND X (s DELETE e))))``,

  REPEAT STRIP_TAC THEN1
  SIMP_TAC (srw_ss()) [ITSET_THM, nR_AND_def, R_NEVER_def] THEN
  SIMP_TAC (srw_ss()) [nR_AND_def, R_NEVER_def] THEN
  Q.ABBREV_TAC `g = \e acc. R_AND (X e) acc` THEN
  Q_TAC SUFF_TAC `pred_set$ITSET g (e INSERT s) (\s. PosInf) =
                  g e (pred_set$ITSET g (s DELETE e) (\s. PosInf))` THEN1 SRW_TAC [][Abbr`g`] THEN
  MATCH_MP_TAC COMMUTING_ITSET_RECURSES THEN
SRW_TAC [][Abbr`g`] THEN METIS_TAC[R_AND_COMM, R_AND_ASSOC]);


val nR_AND_DRBD1 = store_thm("nR_AND_DRBD1",
  ``!p X t s.
       FINITE s ==>
       (DRBD_event p (nR_AND X s) t =
       (DRBD_series (rv_to_event p X t) s) INTER p_space p) ``,
NTAC 3 GEN_TAC 
THEN SET_INDUCT_TAC
     THEN1(RW_TAC std_ss[nR_AND_def, R_NEVER_def]
        THEN ONCE_REWRITE_TAC[pred_setTheory.ITSET_def]
        THEN RW_TAC std_ss[]
          THEN1(RW_TAC std_ss[DRBD_series_def, rv_to_event_def]
	        THEN `{DRBD_event p (X i) t | i IN {}}  = {}` by ASM_SET_TAC[]
		THEN RW_TAC std_ss[BIGINTER_EMPTY, DRBD_event_def]
		THEN ASM_SET_TAC[lt_infty])
     THEN ASM_SET_TAC[FINITE_EMPTY])
THEN RW_TAC std_ss[nR_AND_THM]
THEN `s DELETE (e:num) = s`  by ASM_SET_TAC[]
THEN ONCE_ASM_REWRITE_TAC[]
THEN RW_TAC std_ss[R_AND_INTER]
THEN RW_TAC std_ss[DRBD_series_def, rv_to_event_def]
THEN `(e:num) INSERT s = {e} UNION s` by ASM_SET_TAC[]
THEN ONCE_ASM_REWRITE_TAC[]
THEN  `{DRBD_event p (X i) t | i IN {e} UNION s} =
       {DRBD_event p (X e) t} UNION {DRBD_event p (X i) t | i IN  s}` by ASM_SET_TAC[]
THEN RW_TAC std_ss[BIGINTER_UNION, BIGINTER_SING]
THEN RW_TAC std_ss[INTER_ASSOC]);


val nR_AND_DRBD_series = store_thm("nR_AND_DRBD_series",
  `` !p X t s.
            FINITE s /\ s <> {} ==>
            (DRBD_event p (nR_AND X s) t = DRBD_series (rv_to_event p X t) s)  ``,

RW_TAC std_ss[nR_AND_DRBD1]
THEN KNOW_TAC(``?e:num. s = e INSERT (s DELETE e) ``)
     THEN1 ASM_SET_TAC[]
THEN RW_TAC std_ss[]
THEN ONCE_ASM_REWRITE_TAC[]
THEN RW_TAC std_ss[DRBD_series_def, rv_to_event_def]
THEN `(e:num) INSERT (s DELETE e)  = {e} UNION (s DELETE e)` by ASM_SET_TAC[]
THEN ONCE_ASM_REWRITE_TAC[]
THEN `{DRBD_event p (X i) t | i IN {e} UNION (s DELETE e)} =
      {DRBD_event p (X e) t} UNION {DRBD_event p (X i) t | i IN  (s DELETE e)}` by ASM_SET_TAC[]

THEN RW_TAC std_ss[BIGINTER_UNION]
THEN RW_TAC std_ss[BIGINTER_SING]
THEN RW_TAC std_ss[DRBD_event_def]
THEN ASM_SET_TAC[]);


  
val nR_OR_def = Define `nR_OR X s = pred_set$ITSET (\e acc. R_OR(X e) acc) s R_ALWAYS `;

val nR_OR_THM = store_thm("nR_OR_THM",
  `` !X.
            (nR_OR X {} = (\s. 0)) /\
            !e s.
                FINITE s ==>
                (nR_OR X (e INSERT s) = R_OR (X e) (nR_OR X (s DELETE e)))``,

  REPEAT STRIP_TAC THEN1
  SIMP_TAC (srw_ss()) [ITSET_THM, nR_OR_def, R_ALWAYS_def] THEN
  SIMP_TAC (srw_ss()) [nR_OR_def, R_ALWAYS_def] THEN
  Q.ABBREV_TAC `g = \e acc. R_OR (X e) acc` THEN
  Q_TAC SUFF_TAC `pred_set$ITSET g (e INSERT s) (\s. 0) =
                  g e (pred_set$ITSET g (s DELETE e) (\s. 0))` THEN1 SRW_TAC [][Abbr`g`] THEN
  MATCH_MP_TAC COMMUTING_ITSET_RECURSES THEN
  SRW_TAC [][Abbr`g`] THEN METIS_TAC[R_OR_COMM, R_OR_ASSOC]);

val nR_OR_DRBD1 = store_thm("nR_OR_DRBD1",
  ``    !p X t s.
            FINITE s ==>
            0 <= t ==>
            (DRBD_event p (nR_OR X s) t = DRBD_parallel (rv_to_event p X t) s)
        ``,
NTAC 3 GEN_TAC 
THEN SET_INDUCT_TAC
     THEN1(RW_TAC std_ss[nR_OR_def, R_ALWAYS_def]
        THEN ONCE_REWRITE_TAC[pred_setTheory.ITSET_def]
        THEN RW_TAC std_ss[]
          THEN1(RW_TAC std_ss[DRBD_parallel_def, rv_to_event_def]
	        THEN `{DRBD_event p (X i) t | i IN {}}  = {}` by ASM_SET_TAC[]
		THEN RW_TAC std_ss[BIGUNION_EMPTY, DRBD_event_def]
		THEN `0 = Normal 0` by METIS_TAC[extreal_of_num_def]
		THEN ONCE_ASM_REWRITE_TAC[]
		THEN RW_TAC std_ss[extreal_lt_eq]
		THEN `~(t < 0)` by METIS_TAC[real_lte]
		THEN RW_TAC std_ss[]
		THEN ASM_SET_TAC[])
	 THEN ASM_SET_TAC[FINITE_EMPTY]	)
THEN RW_TAC std_ss[]	 
THEN RW_TAC std_ss[nR_OR_THM]
THEN `s DELETE (e:num) = s`  by ASM_SET_TAC[]
THEN ONCE_ASM_REWRITE_TAC[]
THEN RW_TAC std_ss[R_OR_UNION]
THEN RW_TAC std_ss[DRBD_parallel_def, rv_to_event_def]
THEN `(e:num) INSERT s = {e} UNION s` by ASM_SET_TAC[]
THEN ONCE_ASM_REWRITE_TAC[]
THEN  `{DRBD_event p (X i) t | i IN {e} UNION s} =
       {DRBD_event p (X e) t} UNION {DRBD_event p (X i) t | i IN  s}` by ASM_SET_TAC[]
THEN RW_TAC std_ss[BIGUNION_UNION, BIGUNION_SING]);

val nR_OR_DRBD_parallel = store_thm("nR_OR_DRBD_parallel",
  ``!p X t s.
            FINITE s /\ 0 <= t ==>
            (DRBD_event p (nR_OR X s) t = DRBD_parallel (rv_to_event p X t) s)
         ``,

RW_TAC std_ss[nR_OR_DRBD1]);




(******************************************************************************)
(*             Definitions and Theorems ported from Isabelle/HOL or           *)
(*                           HOL Light                                        *)
(*                                                                            *)
(******************************************************************************)

val to_nat_on_def = Define `to_nat_on (A:'a->bool) = 
 @f.   if FINITE A then  ( BIJ f A (count (CARD A ))) else  (BIJ f A univ(:num))`;


val inv_into_def = Define `inv_into s f = (\y. LINV f s y) `;

val from_nat_into_def = Define `from_nat_into A  =

(\n. if  (n IN (IMAGE (to_nat_on A) A)) then (inv_into A (to_nat_on A) n) else   (@s. s IN A)) `;

val range_def = Define `range (f:'a->'b)  =  IMAGE f  univ(:'a)`;


val CARD_EQ_BIJ = store_thm("CARD_EQ_BIJ",
  ``!s t. FINITE s /\ FINITE t /\ (CARD s = CARD t) ==> ?f. BIJ f s t``, 
RW_TAC std_ss[BIJ_DEF,INJ_DEF,SURJ_DEF]

THEN KNOW_TAC(``
        !(f:'a->'b).
	          ((!x. x IN s ==> f x IN t) /\
             !x y. x IN s /\ y IN s ==> (f x = f y) ==> (x = y)) /\
            (!x. x IN s ==> f x IN t) /\
            (!x. x IN t ==> ?y. y IN s /\ (f y = x)) <=>
            (!x. x IN s ==> f x IN t) /\
            (!y. y IN t ==> ?x. x IN s /\ (f x = y)) /\
            !x y. x IN s /\ y IN s /\ (f x = f y) ==> (x = y)``)
      THEN1(RW_TAC std_ss[]
            THEN EQ_TAC
	    THEN RW_TAC std_ss[])
THEN RW_TAC std_ss[CARD_EQ_BIJECTION]);


val to_nat_on_finite = store_thm("to_nat_on_finite",
  ``!A. FINITE A ==> BIJ (to_nat_on A) A (count (CARD A))  ``,
RW_TAC std_ss[]
THEN KNOW_TAC (``FINITE (count (CARD A)) ``)
     THEN1(ASM_SET_TAC[FINITE_COUNT])
THEN DISCH_TAC     
THEN KNOW_TAC(``CARD A = CARD (count (CARD A)) ``)
     THEN1(ASM_SET_TAC[CARD_COUNT])
THEN DISCH_TAC     
THEN KNOW_TAC(``!B. FINITE B ==> ?f. BIJ f B (count (CARD (B:'a->bool))) ``)
     THEN1(RW_TAC std_ss[]
          THEN KNOW_TAC (``FINITE (count (CARD B)) ``)
          THEN1(ASM_SET_TAC[FINITE_COUNT])
          THEN DISCH_TAC     
          THEN KNOW_TAC(``CARD B = CARD (count (CARD B)) ``)
              THEN1(ASM_SET_TAC[CARD_COUNT])
          THEN DISCH_TAC     
	  THEN RW_TAC std_ss[CARD_EQ_BIJ])
THEN RW_TAC std_ss[]	  
THEN RW_TAC std_ss[to_nat_on_def]
THEN SELECT_ELIM_TAC
THEN RW_TAC std_ss[]);



val finite_imageD_lem = store_thm("finite_imageD_lem",
  ``!f A B.
         FINITE B ==>  (B = IMAGE f A) /\ INJ f A (IMAGE f A) ==> FINITE A   ``,
GEN_TAC THEN GEN_TAC THEN SET_INDUCT_TAC
THEN1( RW_TAC std_ss[]
       THEN FULL_SIMP_TAC std_ss[IMAGE_DEF]
       THEN `A = {}` by   ASM_SET_TAC[INJ_EMPTY, FINITE_EMPTY]
       THEN RW_TAC std_ss[]
       THEN RW_TAC std_ss[FINITE_EMPTY])
THEN STRIP_TAC 
THEN KNOW_TAC(``e:'b IN (IMAGE (f:'a->'b) A) ``) 
    THEN1(ASM_SET_TAC[])
THEN DISCH_TAC
THEN KNOW_TAC(``?y. y IN A /\ (e = f y) ``)
     THEN1(FULL_SIMP_TAC std_ss[INJ_DEF]
           THEN ASM_SET_TAC[])
THEN STRIP_TAC
THEN KNOW_TAC(``s = (IMAGE (f:'a->'b) A ) DIFF {e} ``)
     THEN1(ASM_SET_TAC[])
THEN DISCH_TAC
THEN KNOW_TAC(``s = (IMAGE (f:'a->'b)  (A  DIFF {y})) ``)
     THEN1(PSET_TAC[INJ_DEF]
           THEN ASM_SET_TAC[])
THEN DISCH_TAC
THEN KNOW_TAC(``INJ f (A DIFF {y}) (IMAGE f (A DIFF {y})) ``)
     THEN1(` A DIFF {y} =  A DELETE y` by ASM_SET_TAC[]
           THEN ONCE_ASM_REWRITE_TAC[]
	   THEN `IMAGE f ( A DELETE y) = (IMAGE f A )DELETE f y` by ASM_SET_TAC[]
           THEN ONCE_ASM_REWRITE_TAC[]
	   THEN PSET_TAC[INJ_DELETE])
THEN DISCH_TAC
THEN KNOW_TAC(`` FINITE (IMAGE (f:'a->'b) ( A DELETE y)) = FINITE (A DELETE y) ``)
     THEN1(RW_TAC std_ss[]
	   THEN KNOW_TAC(`` (!u w. u IN (A DELETE y)  /\ w IN (A DELETE y) /\
	                           (f u = f w) ==> (u = w))``)
                THEN1(ASM_SET_TAC[INJ_DEF])
	  THEN PSET_TAC [FINITE_IMAGE_INJ_EQ] )
THEN DISCH_TAC	  

THEN RW_TAC std_ss[]
THEN ` A DIFF {y} =  A DELETE y` by ASM_SET_TAC[]
THEN FULL_SIMP_TAC std_ss[]           
THEN FULL_SIMP_TAC std_ss[]
THEN KNOW_TAC(`` FINITE ( A DELETE y)``)
     THEN1(PSET_TAC[])
THEN DISCH_TAC
THEN METIS_TAC[ FINITE_DELETE]);


val finite_imageD = store_thm("finite_imageD",
  `` !f A. FINITE (IMAGE f A) /\ INJ f A (IMAGE f A) ==> FINITE A ``,
RW_TAC std_ss[]
THEN DEP_REWRITE_TAC[finite_imageD_lem]
THEN Q.EXISTS_TAC`f`
THEN Q.EXISTS_TAC` IMAGE f A`
THEN ASM_SET_TAC[finite_imageD_lem]);

val finite_imageD_lem2 = store_thm("finite_imageD_lem2",
  ``! f A t B .  FINITE B ==>  (B = IMAGE f A) /\ INJ f A t ==> FINITE A   ``,
GEN_TAC THEN GEN_TAC THEN GEN_TAC THEN SET_INDUCT_TAC
THEN1( RW_TAC std_ss[]
       THEN FULL_SIMP_TAC std_ss[IMAGE_DEF]
       THEN `A = {}` by   ASM_SET_TAC[INJ_EMPTY, FINITE_EMPTY]
       THEN RW_TAC std_ss[]
       THEN RW_TAC std_ss[FINITE_EMPTY])
THEN STRIP_TAC 
THEN KNOW_TAC(``e:'b IN (IMAGE (f:'a->'b) A) ``) 
    THEN1(ASM_SET_TAC[])
THEN DISCH_TAC
THEN KNOW_TAC(``?y. y IN A /\ (e = f y) ``)
     THEN1(FULL_SIMP_TAC std_ss[INJ_DEF]
           THEN ASM_SET_TAC[])
THEN STRIP_TAC
THEN KNOW_TAC(``s = (IMAGE (f:'a->'b) A ) DIFF {e} ``)
     THEN1(ASM_SET_TAC[])
THEN DISCH_TAC
THEN KNOW_TAC(``s = (IMAGE (f:'a->'b)  (A  DIFF {y})) ``)
     THEN1(PSET_TAC[INJ_DEF]
           THEN ASM_SET_TAC[])
THEN DISCH_TAC
THEN KNOW_TAC(``INJ (f:'a->'b) (A DIFF {y}) (t DELETE f y) ``)
     THEN1(` A DIFF {y} =  A DELETE y` by ASM_SET_TAC[]
           THEN ONCE_ASM_REWRITE_TAC[]
	   THEN PSET_TAC[INJ_DELETE])
THEN DISCH_TAC
THEN KNOW_TAC(`` FINITE (IMAGE (f:'a->'b) ( A DELETE y)) = FINITE (A DELETE y) ``)
     THEN1(RW_TAC std_ss[]
	   THEN KNOW_TAC(`` (!u w. u IN (A DELETE y)  /\ w IN (A DELETE y) /\
	                    (f u = f w) ==> (u = w))``)
                THEN1(ASM_SET_TAC[INJ_DEF])
	  THEN PSET_TAC [FINITE_IMAGE_INJ_EQ] )
THEN DISCH_TAC	  

THEN RW_TAC std_ss[]
THEN ` A DIFF {y} =  A DELETE y` by ASM_SET_TAC[]
THEN FULL_SIMP_TAC std_ss[]           
THEN FULL_SIMP_TAC std_ss[]
THEN KNOW_TAC(`` FINITE ( A DELETE y)``)
     THEN1(PSET_TAC[])
THEN DISCH_TAC
THEN METIS_TAC[ FINITE_DELETE]);


val finite_imageD2 = store_thm("finite_imageD2",
  `` !f A t. FINITE (IMAGE f A) /\ INJ f A t==> FINITE A ``,
RW_TAC std_ss[]
THEN DEP_REWRITE_TAC[finite_imageD_lem2]
THEN Q.EXISTS_TAC`f`
THEN Q.EXISTS_TAC `t`
THEN Q.EXISTS_TAC` IMAGE f A`
THEN ASM_SET_TAC[finite_imageD_lem2]);


val countableE_infinite = store_thm("countableE_infinite",
  ``!A. countable A /\ INFINITE A ==> ?f. BIJ f A univ(:num)  ``,

RW_TAC std_ss[]
THEN KNOW_TAC(``!B:'a->bool. countable B ==> ?f. INJ f B univ(:num) ``)
     THEN1(RW_TAC std_ss[countable_def])
THEN RW_TAC std_ss[]     
THEN `?k. INJ k A univ(:num)` by RW_TAC std_ss[]
THEN KNOW_TAC(``INFINITE (IMAGE (k:'a->num) A) ``)
    THEN1(KNOW_TAC(``FINITE (IMAGE (k:'a->num) A) /\ INJ k A univ(:num) ==> FINITE A ``)
    	  THEN1( METIS_TAC[finite_imageD2])
          THEN DISCH_TAC	  
	  THEN FULL_SIMP_TAC std_ss[CONJ_COMM]
	  THEN `FINITE (IMAGE (k:'a->num) A) ==> FINITE A` by RW_TAC std_ss[]
	  THEN UNDISCH_TAC(``INFINITE (A:'a->bool) ``)
	  THEN METIS_TAC[MONO_NOT_EQ, CONJ_ASSOC, CONJ_COMM])
THEN DISCH_TAC	  
THEN KNOW_TAC(``BIJ (LINV (enumerate (IMAGE k A)) univ(:num) ) (IMAGE (k:'a->num) A) univ(:num) ``)
     THEN1(KNOW_TAC(``BIJ (enumerate (IMAGE k A)) univ(:num) (IMAGE (k:'a->num) A)``)
           THEN1(KNOW_TAC(``countable (IMAGE (k:'a->num) A) ``)
	         THEN1( RW_TAC std_ss[image_countable])
		 THEN DISCH_TAC THEN ASM_SET_TAC[COUNTABLE_ALT_BIJ])
           THEN DISCH_TAC	   
	   THEN ASM_SET_TAC[BIJ_LINV_BIJ])
THEN DISCH_TAC	   
THEN KNOW_TAC(``BIJ (k:'a->num) A (IMAGE k A) ``)
     THEN1(METIS_TAC[INJ_IMAGE_BIJ])
THEN DISCH_TAC
THEN KNOW_TAC(``BIJ ((LINV (enumerate (IMAGE (k:'a->num) A)) univ(:num)) o k) A univ(:num) ``)
     THEN1(DEP_REWRITE_TAC[BIJ_COMPOSE]
           THEN Q.EXISTS_TAC `(IMAGE (k:'a->num) A)`
	   THEN RW_TAC std_ss[])
THEN RW_TAC std_ss[]	   
THEN METIS_TAC[]);

val to_nat_on_infinite = store_thm("to_nat_on_infinite",
  ``!A. countable A /\ INFINITE A ==> BIJ (to_nat_on A) A univ(:num)  ``,
RW_TAC std_ss[to_nat_on_def]
THEN SELECT_ELIM_TAC
THEN RW_TAC std_ss[]
THEN RW_TAC std_ss[countableE_infinite]);

val BIJ_CONG = store_thm("BIJ_CONG",
  ``!s t f g. (!a. a IN s ==> (f a = g a)) ==> (BIJ f s t =  BIJ g s t)    ``,

RW_TAC std_ss[BIJ_DEF, INJ_DEF, SURJ_DEF]
THEN EQ_TAC
THEN RW_TAC std_ss[]
THEN POP_ASSUM MP_TAC THEN POP_ASSUM (MP_TAC o Q.SPEC `x`)
THEN RW_TAC std_ss[]
THEN FULL_SIMP_TAC std_ss[]
THEN Q.EXISTS_TAC `y`
THEN METIS_TAC[]);


val inv_into_f_f = store_thm("inv_into_f_f",
  ``!A f x t. INJ f A t /\ x IN A ==> (inv_into A f (f x ) = x)   ``,

RW_TAC std_ss[inv_into_def]
THEN METIS_TAC[LINV_DEF]);



val from_nat_into_nat_on = store_thm("from_nat_into_nat_on",
  `` !A a. countable A /\ a IN A ==> (from_nat_into A (to_nat_on A a) = a)``,

 RW_TAC std_ss[from_nat_into_def]
 THEN1(Cases_on `FINITE A`
       THEN1(KNOW_TAC(``INJ (to_nat_on A) A (count (CARD A)) ``)
             THEN1(METIS_TAC[to_nat_on_finite, BIJ_DEF])
	     THEN DISCH_TAC
	     THEN METIS_TAC[inv_into_f_f])
       THEN KNOW_TAC(`` INJ (to_nat_on A) A univ(:num)``)
            THEN1(METIS_TAC [BIJ_DEF, to_nat_on_infinite])
       THEN METIS_TAC[inv_into_f_f])

THEN SELECT_ELIM_TAC
THEN RW_TAC std_ss[]
     THEN1(METIS_TAC[])
THEN KNOW_TAC(``to_nat_on A a IN IMAGE (to_nat_on A) A ``)
    THEN1(ASM_SET_TAC[])
THEN METIS_TAC[] );   

val subset_range_from_nat_into = store_thm("subset_range_from_nat_into",
  ``!A. countable A ==> A SUBSET range (from_nat_into A)  ``,
RW_TAC std_ss[]
THEN RW_TAC std_ss[range_def]
THEN KNOW_TAC(``!a. a IN A ==> (from_nat_into A (to_nat_on  A a) = a)   ``)
     THEN1(RW_TAC std_ss[from_nat_into_nat_on])
THEN DISCH_TAC     
THEN ASM_SET_TAC[]);


val inv_into_into = store_thm("inv_into_into",
  ``!x f A. x IN IMAGE f A ==> inv_into A f x IN A  ``,
RW_TAC std_ss[inv_into_def]
THEN RW_TAC std_ss[LINV_LO]
THEN RW_TAC std_ss[LINV_OPT_def]
THEN SELECT_ELIM_TAC
THEN RW_TAC std_ss[]
THEN ASM_SET_TAC[]);

val from_nat_into_lem = store_thm("from_nat_into_lem",
  ``!A n. A <>{} ==> from_nat_into A n IN A ``,
RW_TAC std_ss[from_nat_into_def]
    THEN1(METIS_TAC[inv_into_into])
THEN SELECT_ELIM_TAC    
THEN RW_TAC std_ss[]
THEN ASM_SET_TAC[]);


val range_from_nat_into_subset = store_thm("range_from_nat_into_subset",
  ``!A. A <> {} ==> range (from_nat_into A ) SUBSET A ``,
RW_TAC std_ss[]
THEN RW_TAC std_ss[range_def]
THEN PSET_TAC[from_nat_into_lem]);

val range_from_nat_into = store_thm("range_from_nat_into",
  ``!A. countable A /\ A <> {} ==> (range (from_nat_into A) = A)  ``,

PSET_TAC[range_from_nat_into_subset, subset_range_from_nat_into, SUBSET_ANTISYM]);

val UN_from_nat_into = store_thm("UN_from_nat_into",
  ``!s A.
        s<> {} /\ countable (s:num->bool) ==>
	(BIGUNION { A i| i IN s} =
	BIGUNION {A (from_nat_into s i)| i IN univ(:num)})``,

RW_TAC std_ss[]
THEN KNOW_TAC(`` BIGUNION {A (from_nat_into (s:num->bool) i) | i IN univ(:num)} =
                 BIGUNION (IMAGE A (range (from_nat_into s))) ``)

THEN1(RW_TAC std_ss[IMAGE_DEF, range_def]
      THEN PSET_TAC[]
      THEN ASM_SET_TAC[IMAGE_DEF])
THEN RW_TAC std_ss[]
THEN RW_TAC std_ss[range_def]
THEN KNOW_TAC(``IMAGE (from_nat_into s) univ(:num) = s:num->bool ``)
     THEN1(RW_TAC std_ss[GSYM range_def]
           THEN PSET_TAC[range_from_nat_into])
THEN DISCH_TAC THEN ONCE_ASM_REWRITE_TAC[]	   
THEN ASM_SET_TAC[]);

val sigma_sets_union_1 = store_thm ("sigma_sets_union_1",
 ``!X A B.
        countable B /\  (!b. b IN B ==> b IN sigma_sets X A) ==>
        (BIGUNION B IN sigma_sets X A)``,
RW_TAC std_ss[]
THEN Cases_on `B={}`
THEN1(RW_TAC std_ss[BIGUNION_EMPTY]
      THEN RW_TAC std_ss[sigma_sets_empty])
THEN `!n. from_nat_into B n IN B` by RW_TAC std_ss[from_nat_into_lem]       
THEN `range(from_nat_into B) = B` by RW_TAC std_ss[range_from_nat_into]
THEN KNOW_TAC(`` !i. (from_nat_into B) i IN sigma_sets X A ==>
                     BIGUNION {from_nat_into B i| i IN univ(:num)} IN sigma_sets X A``)
     THEN1 METIS_TAC[sigma_sets_union]
THEN DISCH_TAC     
THEN KNOW_TAC(``!i. from_nat_into B i IN sigma_sets X A  ``)
     THEN1(RW_TAC std_ss[])
THEN DISCH_TAC     
THEN FULL_SIMP_TAC std_ss[]
THEN KNOW_TAC(``  BIGUNION B= BIGUNION {from_nat_into B i | i IN univ(:num)} ``)
    THEN1(AP_TERM_TAC
          THEN RW_TAC std_ss[GSYM IMAGE_DEF]
	  THEN `(\i. from_nat_into B i) = from_nat_into B` by METIS_TAC[ETA_AX]
	  THEN RW_TAC std_ss[GSYM range_def]
	  THEN PSET_TAC [range_from_nat_into])
THEN RW_TAC std_ss[]);	  

(*========================================================================*)





(******************************************************************************)
(*                                                                            *)
(*                  additional required properties of indep_sets              *)
(*                                                                            *)
(******************************************************************************)


val INT_from_nat_into = store_thm("INT_from_nat_into",
  ``!s A.
        s<> {} /\ countable (s:num->bool) ==>
        (BIGINTER { A i| i IN s} =
	 BIGINTER {A (from_nat_into s i)| i IN univ(:num)})``,

RW_TAC std_ss[]
THEN KNOW_TAC(`` BIGINTER {A (from_nat_into (s:num->bool) i) | i IN univ(:num)} =
                 BIGINTER (IMAGE A (range (from_nat_into s))) ``)

THEN1(RW_TAC std_ss[IMAGE_DEF, range_def]
      THEN PSET_TAC[]
      THEN ASM_SET_TAC[IMAGE_DEF])
THEN RW_TAC std_ss[]
THEN RW_TAC std_ss[range_def]
THEN KNOW_TAC(``IMAGE (from_nat_into s) univ(:num) = s:num->bool ``)
     THEN1(RW_TAC std_ss[GSYM range_def]
           THEN PSET_TAC[range_from_nat_into])
THEN DISCH_TAC THEN ONCE_ASM_REWRITE_TAC[]	   
THEN ASM_SET_TAC[]);



val indep_sets_sigma_1 = store_thm("indep_sets_sigma_1",
  ``!p X s. indep_sets p (\i. {X i})  s ==>
             indep_sets p (\i. sigma_sets(p_space p) {X i}) s``,
RW_TAC std_ss[]
THEN DEP_REWRITE_TAC[indep_sets_sigma']
THEN RW_TAC std_ss[]
     THEN1(FULL_SIMP_TAC std_ss[indep_sets])
THEN RW_TAC std_ss[Int_stable]
THEN PSET_TAC[]);



val indep_sets_collect_sigma_BIGUNION_events = store_thm(
  "indep_sets_collect_sigma_BIGUNION_events",
  ``  !p s J Y.
            indep_sets p (\i. {Y i}) (BIGUNION {s j | j IN J}) /\ J <> {} /\
            (!i. i IN J ==> countable (s i)) /\ FINITE J /\
            disjoint_family_on s J ==>
            indep_sets p (\j. {BIGUNION {Y i | i IN s j}}) J``,

RW_TAC std_ss[]
THEN `prob_space p` by FULL_SIMP_TAC std_ss[indep_sets]
THEN  KNOW_TAC(`` indep_sets p
             (\j. sigma_sets (m_space p) (BIGUNION { {Y (i:num)} | i IN s j})) J``)

          THEN1(DEP_REWRITE_TAC [indep_sets_collect_sigma]
                THEN RW_TAC std_ss[]
                THEN RW_TAC std_ss[Int_stable]
		THEN PSET_TAC[])
THEN RW_TAC std_ss[]	
THEN `indep_sets p (\j. BIGUNION {{Y i} | i IN s j}) J` by RW_TAC std_ss[indep_sets_collect_sigma_BIGUNION]
THEN KNOW_TAC(``!j. j IN (J:num->bool) ==>
                   {BIGUNION {(Y:num->'a->bool) i | i IN s j}} SUBSET
		   sigma_sets (m_space p) (BIGUNION {{Y i} | i IN s j})``)
     THEN1(RW_TAC std_ss[]
           THEN KNOW_TAC(``BIGUNION {(Y:num->'a->bool) i | i IN s (j:num)} IN
	                   sigma_sets (m_space p) (BIGUNION {{Y i} | i IN s j}) ``)

               THEN1(KNOW_TAC(``!i. i IN s (j:num) ==>
	                            (Y:num->'a->bool) i IN
				    sigma_sets (m_space p) (BIGUNION {{Y i} | i IN s j}) ``)
                     THEN1(RW_TAC std_ss[]
		           THEN KNOW_TAC(``Y i IN BIGUNION {{(Y:num->'a->bool) i} | i IN s (j:num)}``)
			   THEN1(PSET_TAC[]
			         THEN Q.EXISTS_TAC` {Y i}`
				 THEN ASM_SET_TAC[])
			   THEN ASM_SET_TAC[sigma_sets_basic])
		    THEN RW_TAC std_ss[]
		    THEN KNOW_TAC(``!b. b IN {(Y:num->'a->bool) i | i IN s (j:num)} ==>
		                        b IN sigma_sets (m_space p) (BIGUNION {{Y i} | i IN s j})  ``)
		        THEN1( ASM_SET_TAC[])
		    THEN DISCH_TAC
		    THEN DEP_REWRITE_TAC[sigma_sets_union_1]
		    THEN RW_TAC std_ss[]
		    THEN RW_TAC std_ss[GSYM IMAGE_DEF]
		    THEN `countable (s (j:num))` by RW_TAC std_ss[]
                    THEN ASM_SET_TAC[image_countable])
              THEN ASM_SET_TAC[])
THEN RW_TAC std_ss[]
THEN DEP_REWRITE_TAC[indep_sets_mono_sets]
THEN Q.EXISTS_TAC`\j. sigma_sets (m_space p) (BIGUNION {{(Y:num->'a->bool) i} | i IN s (j:num)})`
THEN ASM_SET_TAC[]);



val sigma_sets_Inter1 = store_thm ("sigma_sets_Inter1",
 ``!sp st A. 
          (!i. (A:num->'a->bool) i IN sigma_sets sp st /\ A i SUBSET sp) ==>
          BIGINTER {A i | i IN UNIV} IN sigma_sets sp st``,
  REPEAT STRIP_TAC THEN
  KNOW_TAC ``(!i:num. A i IN sigma_sets sp st) ==>
             (!i:num. sp DIFF A i IN sigma_sets sp st)`` THENL
  [METIS_TAC [sigma_sets_compl], DISCH_TAC] THEN
  KNOW_TAC ``BIGUNION {sp DIFF A i | (i:num) IN UNIV} IN sigma_sets sp st`` THENL
  [ONCE_REWRITE_TAC [METIS [] ``sp DIFF A i = (\i. sp DIFF A i) i``] THEN
   MATCH_MP_TAC sigma_sets_union THEN METIS_TAC [], DISCH_TAC] THEN
  KNOW_TAC
   (``sp DIFF BIGUNION {sp DIFF A i | (i:num) IN UNIV} IN sigma_sets sp st``) THENL
  [MATCH_MP_TAC sigma_sets_compl THEN METIS_TAC [], DISCH_TAC] THEN
  KNOW_TAC ``sp DIFF BIGUNION {sp DIFF A i | i IN UNIV} =
             BIGINTER {A i | (i:num) IN UNIV}`` THENL
  [ALL_TAC, METIS_TAC[]] THEN
  SIMP_TAC std_ss [EXTENSION] THEN GEN_TAC THEN EQ_TAC THENL
  [SIMP_TAC std_ss [IN_DIFF, IN_BIGUNION, IN_BIGINTER, GSPECIFICATION] THEN
   RW_TAC std_ss [] THEN POP_ASSUM K_TAC THEN
   POP_ASSUM (MP_TAC o SPEC ``sp DIFF (A:num->'a->bool) i``) THEN
   ASM_SET_TAC [], ALL_TAC] THEN
  SIMP_TAC std_ss [IN_BIGINTER, IN_DIFF, IN_BIGUNION, GSPECIFICATION] THEN
  RW_TAC std_ss [IN_UNIV] THENL
  [FIRST_X_ASSUM (MP_TAC o SPEC ``(A:num->'a->bool) i``) THEN
   KNOW_TAC ``(?i'. A i = (A:num->'a->bool) i')`` THENL
   [METIS_TAC [], DISCH_TAC THEN ASM_REWRITE_TAC []] THEN
   SPEC_TAC (``x``,``x``) THEN REWRITE_TAC [GSYM SUBSET_DEF] THEN
  
   METIS_TAC [], ALL_TAC] THEN
  ASM_CASES_TAC ``x NOTIN s`` THEN FULL_SIMP_TAC std_ss [] THEN
  RW_TAC std_ss [EXTENSION] THEN EXISTS_TAC ``x`` THEN
  ASM_SIMP_TAC std_ss [IN_DIFF] THEN DISJ2_TAC THEN
  FIRST_X_ASSUM MATCH_MP_TAC THEN METIS_TAC []);


val sigma_sets_INTER_1 = store_thm ("sigma_sets_INTER_1",
 ``!X A B . countable B /\ B <> {} /\ (!b. b IN B ==>
            b IN sigma_sets X A /\ b SUBSET X) ==>
	    (BIGINTER B IN sigma_sets X A)``,
RW_TAC std_ss[]
THEN `!n. from_nat_into B n IN B` by RW_TAC std_ss[from_nat_into_lem]       
THEN `range(from_nat_into B) = B` by RW_TAC std_ss[range_from_nat_into]
THEN KNOW_TAC(`` !i. (from_nat_into B) i IN sigma_sets X A /\
                     (from_nat_into B) i SUBSET X ==>
		     BIGINTER {(from_nat_into B) i| i IN univ(:num)} IN sigma_sets X A``)
     THEN1 METIS_TAC[sigma_sets_Inter1]
    
THEN DISCH_TAC     
THEN KNOW_TAC(``!i. from_nat_into B i IN sigma_sets X A  ``)
     THEN1(RW_TAC std_ss[])
THEN DISCH_TAC     
THEN FULL_SIMP_TAC std_ss[]
THEN KNOW_TAC(``  BIGINTER B= BIGINTER {from_nat_into B i | i IN univ(:num)} ``)
    THEN1(AP_TERM_TAC
          THEN RW_TAC std_ss[GSYM IMAGE_DEF]
	  THEN `(\i. from_nat_into B i) = from_nat_into B` by METIS_TAC[ETA_AX]
	  THEN RW_TAC std_ss[GSYM range_def]
	  THEN PSET_TAC [range_from_nat_into])
THEN RW_TAC std_ss[]);	  


val indep_sets_collect_sigma_BIGINTER_events = store_thm(
  "indep_sets_collect_sigma_BIGINTER_events",
  `` !p s J Y.
            
            indep_sets p (\i. {Y i}) (BIGUNION {s j | j IN J}) /\ J <> {} /\
            (!i. i IN J ==> countable (s i) /\ s i <> {}) /\ FINITE J /\
            disjoint_family_on s J /\
            (!i. i IN BIGUNION {s j | j IN J} ==> Y i SUBSET m_space p) ==>
            indep_sets p (\j. {BIGINTER {Y i | i IN s j}}) J``,

RW_TAC std_ss[]
THEN `prob_space p` by FULL_SIMP_TAC std_ss[indep_sets]
THEN  KNOW_TAC(`` indep_sets p
             (\j. sigma_sets (m_space p) (BIGUNION { {Y (i:num)} | i IN s j})) J``)

          THEN1(DEP_REWRITE_TAC [indep_sets_collect_sigma]
                THEN RW_TAC std_ss[]
                THEN RW_TAC std_ss[Int_stable]
		THEN PSET_TAC[])
THEN RW_TAC std_ss[]	
THEN `indep_sets p (\j. BIGUNION {{Y i} | i IN s j}) J` by RW_TAC std_ss[indep_sets_collect_sigma_BIGUNION]
THEN KNOW_TAC(``!j. j IN (J:num->bool) ==>
                    {BIGINTER {(Y:num->'a->bool) i | i IN s j}} SUBSET
		    sigma_sets (m_space p) (BIGUNION {{Y i} | i IN s j})``)
     THEN1(RW_TAC std_ss[]
           THEN KNOW_TAC(``BIGINTER {(Y:num->'a->bool) i | i IN s (j:num)} IN
	                   sigma_sets (m_space p) (BIGUNION {{Y i} | i IN s j}) ``)

               THEN1(KNOW_TAC(``!i. i IN s (j:num) ==>
	                            (Y:num->'a->bool) i IN
				    sigma_sets (m_space p) (BIGUNION {{Y i} | i IN s j}) /\ Y i SUBSET (m_space p) ``)
                     THEN1(GEN_TAC THEN DISCH_TAC
		           THEN KNOW_TAC(``Y i IN BIGUNION {{(Y:num->'a->bool) i} | i IN s (j:num)}``)
			   THEN1(PSET_TAC[]
			         THEN Q.EXISTS_TAC` {Y i}`
				 THEN ASM_SET_TAC[])
			    THEN RW_TAC std_ss[] THEN1 ASM_SET_TAC[sigma_sets_basic]
			    THEN `i IN BIGUNION {s j | j IN J}` by ASM_SET_TAC[]
			    THEN ASM_SET_TAC[])
		    THEN RW_TAC std_ss[]
		    THEN KNOW_TAC(``!b. b IN {(Y:num->'a->bool) i | i IN s (j:num)} ==>
		                        b IN sigma_sets (m_space p) (BIGUNION {{Y i} | i IN s j})  ``)
		        THEN1( ASM_SET_TAC[])
		    THEN DISCH_TAC
		    THEN DEP_REWRITE_TAC[sigma_sets_INTER_1]
		    THEN RW_TAC std_ss[]
                         THEN1(RW_TAC std_ss[GSYM IMAGE_DEF]
			      THEN `countable (s (j:num))` by RW_TAC std_ss[]
                    	      THEN ASM_SET_TAC[image_countable])
			 THEN1(ASM_SET_TAC[])
		     THEN ASM_SET_TAC[])	 
              THEN ASM_SET_TAC[])
THEN RW_TAC std_ss[]
THEN DEP_REWRITE_TAC[indep_sets_mono_sets]
THEN Q.EXISTS_TAC`\j. sigma_sets (m_space p) (BIGUNION {{(Y:num->'a->bool) i} | i IN s (j:num)})`
THEN ASM_SET_TAC[]);


(* ========================================================================== *)
(*        Probability of series-parallel, parallel-series                     *)
(*           and nested series-parallel                                       *)
(*                                                                            *)
(* ========================================================================== *)

 val PROB_DRBD_series_parallel = store_thm("PROB_DRBD_series_parallel",
  `` !p X s J.
            indep_sets p (\i. {X i}) (BIGUNION {s j | j IN J}) /\
            (!i. i IN J ==> countable (s i) /\ s i <> {} /\ FINITE (s i)) /\
            FINITE J /\ disjoint_family_on s J /\ J <> {} ==>
            (prob p (DRBD_series  (\j. DRBD_parallel  X (s j)) J) =
             Normal
               (product J
                  (\j.
                       real
                         (1 -
                          Normal
                            (product (s j) (\i. real (1 - prob p (X i))))))))``,

RW_TAC std_ss[]
THEN `prob_space p` by FULL_SIMP_TAC std_ss[indep_sets]
THEN DEP_REWRITE_TAC[PROB_DRBD_series]
THEN RW_TAC std_ss[]
     THEN1(RW_TAC std_ss[DRBD_parallel_def]
           THEN DEP_REWRITE_TAC[indep_sets_collect_sigma_BIGUNION_events]
           THEN RW_TAC std_ss[])

THEN KNOW_TAC (``!j. j IN J ==>
                    (real (prob p (DRBD_parallel (X:num->'a->bool) (s (j:num)))) =
		     real (1 - Normal (product (s j) (\i. real (1 - prob p (X i))))))  ``)
     THEN1(RW_TAC std_ss[]
          THEN AP_TERM_TAC     
 	  THEN DEP_REWRITE_TAC[PROB_DRBD_parallel]
          THEN RW_TAC std_ss[]
               THEN1 (KNOW_TAC(``s (j:num) SUBSET BIGUNION {(s:num->num->bool) j | j IN J} ``)
                      THEN1(ASM_SET_TAC[])
	  	      THEN METIS_TAC[indep_sets_mono_index])
	  THEN FULL_SIMP_TAC std_ss[indep_sets]
	  THEN ASM_SET_TAC[])
THEN RW_TAC std_ss[th]);	  


val PROB_DRBD_series_parallel_rv = store_thm("PROB_DRBD_series_parallel_rv",
  `` !p X t s J.
            indep_sets p (\i. {rv_to_event p X t i})
              (BIGUNION {s j | j IN J}) /\
            (!i. i IN J ==> countable (s i) /\ s i <> {} /\ FINITE (s i)) /\
            FINITE J /\ disjoint_family_on s J /\ J <> {} ==>
            (prob p
               (DRBD_series
                  (\j. DRBD_parallel  (rv_to_event p X t) (s j)) J) =
             Normal
               (product J
                  (\j.
                       real
                         (1 -
                          Normal
                            (product (s j) (\i. real (1 - Rel p (X i) t)))))))``,
RW_TAC std_ss[Rel_def, rv_to_event_def]
THEN DEP_REWRITE_TAC[PROB_DRBD_series_parallel]
THEN RW_TAC std_ss[]);



 val normal_real_product = store_thm("normal_real_product",
  ``   !f J s.
            Normal
              (product J
                 (\j. real (1 - Normal (product (s j) (\i. real (1 - f i)))))) =
            Normal (product J (\j. 1 - product (s j) (\i. real (1 - f i))))``,

RW_TAC std_ss[]
THEN KNOW_TAC(``!j. j IN J==>
                (real (1 - Normal (product (s j) (\i. real (1 - (f:'a->extreal) i)))) =
                (1 - product (s j) (\i. real (1 - f i))))``)

    THEN1(RW_TAC std_ss[]
          THEN KNOW_TAC(``1 = Normal 1 ``)
	       THEN1(METIS_TAC[extreal_of_num_def])
	  THEN RW_TAC std_ss[]
	  THEN RW_TAC std_ss[extreal_sub_def]
	  THEN RW_TAC std_ss[real_normal])
THEN RW_TAC std_ss[th]);	  


 val PROB_DRBD_series_parallel_red = store_thm("PROB_DRBD_series_parallel_red",
  `` !p X s J.
            indep_sets p (\i. {X i}) (BIGUNION {s j | j IN J}) /\
            (!i. i IN J ==> countable (s i) /\ s i <> {} /\ FINITE (s i)) /\
            FINITE J /\ disjoint_family_on s J /\ J <> {} ==>
            (prob p (DRBD_series (\j. DRBD_parallel  X (s j)) J) =
             Normal
               (product J
                  (\j. 1 - (product (s j) (\i. real (1 - prob p (X i)))))))``,
RW_TAC std_ss[normal_real_product, PROB_DRBD_series_parallel]);


val PROB_DRBD_series_parallel_rv_red = store_thm("PROB_DRBD_series_parallel_rv_red",
  ``  !p X t s J.
            indep_sets p (\i. {rv_to_event p X t i})
              (BIGUNION {s j | j IN J}) /\
            (!i. i IN J ==> s i <> {} /\ FINITE (s i)) /\ FINITE J /\
            J <> {} /\ disjoint_family_on s J ==>
            (prob p
               (DRBD_series (\j. DRBD_parallel (rv_to_event p X t) (s j)) J) =
             Normal
               (product J
                  (\j. 1 - product (s j) (\i. real (1 - Rel p (X i) t)))))``,
RW_TAC std_ss[]
THEN KNOW_TAC(``!i:num. i IN J ==> countable ((s:num->num->bool) i)``) 
     THEN1(RW_TAC std_ss[finite_countable])
THEN RW_TAC std_ss[PROB_DRBD_series_parallel_rv, normal_real_product]);



val DRBD_series_IN_events_p = store_thm("DRBD_series_IN_events_p",
  ``!p X s.
            prob_space p /\ (!i. i IN s ==> X i IN events p) /\
            countable s /\ s <> {} ==>
            DRBD_series X s IN events p  ``,

RW_TAC std_ss[DRBD_series_def]
THEN DEP_REWRITE_TAC [EVENTS_COUNTABLE_INTER]
THEN RW_TAC std_ss[]
     THEN1(ASM_SET_TAC[])
     THEN1(RW_TAC std_ss[GSYM IMAGE_DEF]
          THEN METIS_TAC[image_countable])
THEN ASM_SET_TAC[]);	  

val DRBD_parallel_IN_events_p = store_thm("DRBD_parallel_IN_events_p",
  ``!p X s.
            prob_space p /\ FINITE s /\ (!i. i IN s ==> X i IN events p) ==>
            DRBD_parallel X s IN events p``,

RW_TAC std_ss[DRBD_parallel_def]
THEN METIS_TAC[BIGUNION_IN_events_p]);


 val PROB_DRBD_parallel_series = store_thm("PROB_DRBD_parallel_series",
  ``!p X s J.
            indep_sets p (\i. {X i}) (BIGUNION {s j | j IN J}) /\
            (!i. i IN J ==> countable (s i) /\ s i <> {} /\ FINITE (s i)) /\
            FINITE J /\
            disjoint_family_on s J /\ J <> {} ==>
            (prob p (DRBD_parallel (\j. DRBD_series X (s j)) J) =
             1 -
             Normal
               (product J
                  (\j.
                       real
                         (1 -
                          Normal (product (s j) (\i. real (prob p (X i))))))))``,

RW_TAC std_ss[]
THEN KNOW_TAC(`` (!i:num. i IN BIGUNION {s j | j:num IN J} ==> X i IN events p)  ``)
     THEN1(FULL_SIMP_TAC std_ss[indep_sets]
           THEN ASM_SET_TAC[])
THEN RW_TAC std_ss[]	   
THEN `prob_space p` by FULL_SIMP_TAC std_ss[indep_sets]
THEN DEP_REWRITE_TAC[PROB_DRBD_parallel]
THEN RW_TAC std_ss[]
     THEN1(RW_TAC std_ss[DRBD_series_def]
           THEN DEP_REWRITE_TAC[indep_sets_collect_sigma_BIGINTER_events]
           THEN RW_TAC std_ss[]
	   THEN `measure_space p` by FULL_SIMP_TAC std_ss[prob_space_def]
	   THEN METIS_TAC[MEASURABLE_SETS_SUBSET_SPACE, events_def])
THEN NTAC 2 (AP_TERM_TAC)
THEN KNOW_TAC(``!j:num. j IN J ==>
                        (real (1 - prob p (DRBD_series  (X:num->'a->bool) (s j))) =
                         real (1 - Normal (product (s j) (\i. real (prob p (X i))))))``)
     THEN1(RW_TAC std_ss[]
           THEN NTAC 2(AP_TERM_TAC)
	   THEN DEP_REWRITE_TAC[PROB_DRBD_series]
	   THEN RW_TAC std_ss[]
           THEN DEP_REWRITE_TAC[indep_sets_mono_index]
	   THEN Q.EXISTS_TAC`BIGUNION{s j| j IN J}`
	   THEN RW_TAC std_ss[]
	   THEN ASM_SET_TAC[])
THEN RW_TAC std_ss[th]);	   


 val PROB_DRBD_parallel_series_rv = store_thm("PROB_DRBD_parallel_series_rv",
  `` !p X t s J.
            indep_sets p (\i. {rv_to_event p X t i})
              (BIGUNION {s j | j IN J}) /\
            (!i. i IN J ==> countable (s i) /\ s i <> {} /\ FINITE (s i)) /\
            FINITE J /\
            disjoint_family_on s J /\ J <> {} ==>
            (prob p
               (DRBD_parallel (\j. DRBD_series (rv_to_event p X t) (s j)) J) =
             1 -
             Normal
               (product J
                  (\j.
                       real
                         (1 -
                          Normal (product (s j) (\i. real (Rel p (X i) t)))))))``,


RW_TAC std_ss[Rel_def, rv_to_event_def]
THEN DEP_REWRITE_TAC[PROB_DRBD_parallel_series]
THEN RW_TAC std_ss[]);


 val normal_real_product_1 = store_thm("normal_real_product_1",
  `` !f J s.
            1 -
            Normal
              (product J
                 (\j. real (1 - Normal (product (s j) (\i. real (f i)))))) =
            1 - Normal (product J (\j. 1 - product (s j) (\i. real (f i))))``,
RW_TAC std_ss[]
THEN AP_TERM_TAC
THEN RW_TAC std_ss[]
THEN KNOW_TAC(``!j. j IN J==>
                (real (1 - Normal (product (s j) (\i. real ( (f:'a->extreal) i)))) =
                (1 - product (s j) (\i. real ( f i))))``)

    THEN1(RW_TAC std_ss[]
          THEN KNOW_TAC(``1 = Normal 1 ``)
	       THEN1(METIS_TAC[extreal_of_num_def])
	  THEN RW_TAC std_ss[]
	  THEN RW_TAC std_ss[extreal_sub_def]
	  THEN RW_TAC std_ss[real_normal])
THEN RW_TAC std_ss[th]);	  


 val normal_real_product_2 = store_thm("normal_real_product_2",
  ``!f s.
            real (1 - Normal (product s (\i. f i))) =
            1 - product s (\i. f i)``,
RW_TAC std_ss[]
THEN KNOW_TAC(``1 = Normal 1 ``)
     THEN1(METIS_TAC[extreal_of_num_def])
THEN RW_TAC std_ss[]
THEN RW_TAC std_ss[extreal_sub_def, real_normal]);


val PROB_DRBD_parallel_series_red  = store_thm(  "PROB_DRBD_parallel_series_red",
  ``  !p X s J.
            indep_sets p (\i. {X i}) (BIGUNION {s j | j IN J}) /\
            (!i. i IN J ==> countable (s i) /\ s i <> {} /\ FINITE (s i)) /\
            FINITE J /\
            disjoint_family_on s J /\ J <> {} ==>
            (prob p (DRBD_parallel (\j. DRBD_series X (s j)) J) =
             1 -
             Normal
               (product J (\j. 1 - product (s j) (\i. real (prob p (X i))))))``,

RW_TAC std_ss[normal_real_product_1, PROB_DRBD_parallel_series]);

 val PROB_DRBD_parallel_series_rv_red = store_thm("PROB_DRBD_parallel_series_rv_red",
  `` !p X t s J.
            indep_sets p (\i. {rv_to_event p X t i})
              (BIGUNION {s j | j IN J}) /\
            (!i. i IN J ==> s i <> {} /\ FINITE (s i)) /\ FINITE J /\
            J <> {} /\ disjoint_family_on s J ==>
            (prob p
               (DRBD_parallel (\j. DRBD_series (rv_to_event p X t) (s j)) J) =
             1 -
             Normal
               (product J (\j. 1 - product (s j) (\i. real (Rel p (X i) t)))))``,

RW_TAC std_ss[] 
THEN KNOW_TAC(`` (!i:num. i IN J ==> countable ((s:num->num->bool) i))``)
     THEN1(RW_TAC std_ss[finite_countable])
THEN RW_TAC std_ss[PROB_DRBD_parallel_series_rv, normal_real_product_1]);


val nested_BIGUNION_def = Define `
    nested_BIGUNION s L A J =
    BIGUNION
       {s l | l IN BIGUNION {L a | a IN BIGUNION {A j | j IN J}}} `;


val sets_finite_not_empty_def = Define `
    sets_finite_not_empty s L A J =
        (!i. i IN J ==> A i <> {} /\ FINITE (A i)) /\
        (!i.
             i IN BIGUNION {L a | a IN BIGUNION {A j | j IN J}} ==>
             s i <> {} /\ FINITE (s i)) /\
        (!i. i IN BIGUNION {A j | j IN J} ==> L i <> {} /\ FINITE (L i)) /\
        FINITE J /\ disjoint_family_on A J /\ J <> {} /\
        disjoint_family_on s (BIGUNION {L a | a IN BIGUNION {A j | j IN J}}) /\
        disjoint_family_on L (BIGUNION {A j | j IN J})`;


val PROB_DRBD_nested_series_parallel = store_thm("PROB_DRBD_nested_series_parallel",
  `` !p X s L A J.
            indep_sets p (\i. {X i})
              (nested_BIGUNION s L A J) /\
            sets_finite_not_empty s L A J ==>
            (prob p
               (DRBD_series
                  (\j.
                       DRBD_parallel
                         (\a. DRBD_series (\l. DRBD_parallel X (s l)) (L a))
                         (A j)) J) =
             Normal
               (product J
                  (\j.
                       real
                         (1 -
                          Normal
                            (product (A j)
                               (\a.
                                    real
                                      (1 -
                                       Normal
                                         (product (L a)
                                            (\l.
                                                 real
                                                   (1 -
                                                    Normal
                                                      (product (s l)
                                                         (\i.
                                                              real
                                                                (1 -
                                                                 prob p (X i))))))))))))))``,


RW_TAC std_ss[nested_BIGUNION_def, sets_finite_not_empty_def]
THEN KNOW_TAC(``(!i:num.
                 i IN nested_BIGUNION (s:num->num->bool) (L:num->num->bool) (A:num->num->bool) (J:num->bool) ==>
                 (X:num->'a->bool) i IN events p)  ``)
     THEN1(FULL_SIMP_TAC std_ss[indep_sets]
           THEN RW_TAC std_ss[nested_BIGUNION_def]
           THEN ASM_SET_TAC[])
THEN RW_TAC std_ss[nested_BIGUNION_def]	   
THEN `prob_space p` by FULL_SIMP_TAC std_ss[indep_sets]
THEN `!i. FINITE( (A:num->num->bool) i) ==> countable (A i)` by ASM_SET_TAC[finite_countable]
THEN `!i. FINITE( (L:num->num->bool) i) ==> countable (L i)` by ASM_SET_TAC[finite_countable]
THEN `!i. FINITE( (s:num->num->bool) i) ==> countable (s i)` by ASM_SET_TAC[finite_countable]
THEN DEP_REWRITE_TAC[PROB_DRBD_series_parallel]
THEN RW_TAC std_ss[]
     THEN1(RW_TAC std_ss[DRBD_series_def, DRBD_parallel_def]
           THEN DEP_REWRITE_TAC[indep_sets_collect_sigma_BIGINTER_events]
	   THEN RW_TAC std_ss[]
	        THEN1( DEP_REWRITE_TAC[indep_sets_collect_sigma_BIGUNION_events]
		       THEN RW_TAC std_ss[]
		            THEN1(ASM_SET_TAC[])
              	       THEN DEP_REWRITE_TAC[FINITE_BIGUNION]
		       THEN RW_TAC std_ss[]
		            THEN1(RW_TAC std_ss[GSYM IMAGE_DEF]
			          THEN DEP_REWRITE_TAC[IMAGE_FINITE]
				  THEN DEP_REWRITE_TAC[FINITE_BIGUNION]
				  THEN RW_TAC std_ss[]
				       THEN1 (DEP_REWRITE_TAC[IMAGE_FINITE]
				              THEN RW_TAC std_ss[])
				  THEN ASM_SET_TAC[])
		      THEN ASM_SET_TAC[])
	   	THEN1(ASM_SET_TAC[])
		THEN1(DEP_REWRITE_TAC[FINITE_BIGUNION]
		      THEN RW_TAC std_ss[]
		           THEN1(RW_TAC std_ss[GSYM IMAGE_DEF]
			         THEN DEP_REWRITE_TAC[IMAGE_FINITE]
				 THEN RW_TAC std_ss[])
		      THEN ASM_SET_TAC[])
	       THEN KNOW_TAC(``BIGUNION {(X:num->'a->bool) i | i IN (s:num->num->bool) (l:num)} IN
	                       events p``)
	            THEN1(DEP_REWRITE_TAC[BIGUNION_IN_events_p]
		          THEN ASM_SET_TAC[])
	       THEN `measure_space p` by FULL_SIMP_TAC std_ss[prob_space_def]	 	  
	       THEN METIS_TAC[MEASURABLE_SETS_SUBSET_SPACE, events_def])
THEN KNOW_TAC(`` !j:num. j IN (J:num->bool) ==>
           (real
            (1 -
             Normal
               (product ((A:num->num->bool) j)
                  (\a.
                       real
                         (1 -
                          prob p
                            (DRBD_series
                               (\l. DRBD_parallel  (X:num->'a->bool) ((s:num->num->bool) l)) ((L:num->num->bool) a)))))) =
	   real
            (1 -
             Normal
               (product (A j)
                  (\a.
                       real
                         (1 -
                          Normal
                            (product (L a)
                               (\l.
                                    real
                                      (1 -
                                       Normal
                                         (product (s l)
                                            (\i. real (1 - prob p (X i))))))))))))
			       ``)	       

        THEN1(RW_TAC std_ss[]
	      THEN AP_TERM_TAC
	      THEN AP_TERM_TAC
	      THEN AP_TERM_TAC
	      THEN KNOW_TAC(`` !a:num. a IN (A:num->num->bool) j ==>
	   (real
            (1 -
             prob p
               (DRBD_series
	           (\l. DRBD_parallel
		          (X:num->'a->bool) ((s:num->num->bool) l)) (L a))) =
	    real
            (1 -
             Normal
               (product ((L:num->num->bool) a)
                  (\l.
                       real
                         (1 -
                          Normal
                            (product (s l) (\i. real (1 - prob p (X i)))))))))  ``)

		    THEN1(RW_TAC std_ss[]	    	     
		           THEN AP_TERM_TAC
			   THEN AP_TERM_TAC
			   THEN DEP_REWRITE_TAC[PROB_DRBD_series_parallel]
			   THEN RW_TAC std_ss[]
			        THEN1(KNOW_TAC(`` BIGUNION
				                   {(s:num->num->bool) l | l IN (L:num->num->bool) a} SUBSET
						 BIGUNION
						  {s l | l IN BIGUNION
						                {L a | a IN BIGUNION
								              {(A:num->num->bool) j | j IN J}}}``)

                                       THEN1(ASM_SET_TAC[])
		                THEN DISCH_TAC
		    		THEN DEP_REWRITE_TAC[indep_sets_mono_index]
		    		THEN Q.EXISTS_TAC` BIGUNION
				                     {(s:num->num->bool) l | l IN BIGUNION
						                                     {(L:num->num->bool) a | a IN BIGUNION										               {(A:num->num->bool) j | j IN (J:num->bool)}}}` 
		    		THEN RW_TAC std_ss[])
			  THEN ASM_SET_TAC[disjoint_family_on]	)
	     THEN RW_TAC std_ss[]		  
	     THEN RW_TAC std_ss[th])
THEN RW_TAC std_ss[th]);	     


val PROB_DRBD_nested_series_parallel_red = store_thm("PROB_DRBD_nested_series_parallel_red",
  `` !p X s L A J.
             indep_sets p (\i. {X i})
              (nested_BIGUNION s L A J) /\
            sets_finite_not_empty s L A J ==>
            (prob p
               (DRBD_series
                  (\j.
                       DRBD_parallel
                         (\a. DRBD_series (\l. DRBD_parallel X (s l)) (L a))
                         (A j)) J) =
             Normal
               (product J
                  (\j.
                       1 -
                       product (A j)
                         (\a.
                              1 -
                              product (L a)
                                (\l.
                                     1 -
                                     product (s l)
                                       (\i. real (1 - prob p (X i))))))))``,

RW_TAC std_ss[] THEN DEP_REWRITE_TAC[normal_real_product,normal_real_product_1, PROB_DRBD_nested_series_parallel] THEN RW_TAC std_ss[ normal_real_product_2 ]);


val PROB_DRBD_nested_series_parallel_rv = store_thm("PROB_DRBD_nested_series_parallel_rv",
  ``  !p X t s L A J.
            indep_sets p (\i. {rv_to_event p X t i})
              (nested_BIGUNION s L A J) /\
            sets_finite_not_empty s L A J ==>
            (prob p
               (DRBD_series
                  (\j.
                       DRBD_parallel
                         (\a.
                              DRBD_series
                                (\l. DRBD_parallel (rv_to_event p X t) (s l))
                                (L a)) (A j)) J) =
             Normal
               (product J
                  (\j.
                       real
                         (1 -
                          Normal
                            (product (A j)
                               (\a.
                                    real
                                      (1 -
                                       Normal
                                         (product (L a)
                                            (\l.
                                                 real
                                                   (1 -
                                                    Normal
                                                      (product (s l)
                                                         (\i.
                                                              real
                                                                (1 -
                                                                 Rel p (X i)
                                                                   t)))))))))))))``,

RW_TAC std_ss[rv_to_event_def, Rel_def, PROB_DRBD_nested_series_parallel]);


val PROB_DRBD_nested_series_parallel_rv_red = store_thm("PROB_DRBD_nested_series_parallel_rv_red",
  `` !p X t s L A J.
            indep_sets p (\i. {rv_to_event p X t i})
              (nested_BIGUNION s L A J) /\
            sets_finite_not_empty s L A J ==>
            (prob p
               (DRBD_series
                  (\j.
                       DRBD_parallel
                         (\a.
                              DRBD_series
                                (\l. DRBD_parallel (rv_to_event p X t) (s l))
                                (L a)) (A j)) J) =
             Normal
               (product J
                  (\j.
                       1 -
                       product (A j)
                         (\a.
                              1 -
                              product (L a)
                                (\l.
                                     1 -
                                     product (s l)
                                       (\i. real (1 - Rel p (X i) t))))))) ``,

RW_TAC std_ss[rv_to_event_def, Rel_def, PROB_DRBD_nested_series_parallel_red]);


(******************************************************************************)
(*             PROB Spare constructs, AFTER, BEFORE, PAND                     *)
(*                                                                            *)
(*                                                                            *)
(******************************************************************************)

val Rel_R_HSP = store_thm("Rel_R_HSP",
  `` !p t X Y.
          rv_gt0_ninfinity [X; Y] /\
          indep_var p lborel (real o X) lborel (real o Y) ==>
          (Rel p (R_HSP X Y) t =
           1 - (1 - Rel p X t) * (1 - Rel p Y t)) ``,
 RW_TAC std_ss[HSP_OR]
THEN RW_TAC std_ss[Rel_R_OR]);


val Rel_R_CSP = store_thm("Rel_R_CSP",
  ``  !p X Y f_xay f_y f_cond t.
            rv_gt0_ninfinity [X; Y] /\ 0 <= t /\
            (!x y.
                 cond_density lborel lborel p (real o X)
                   (real o Y) y f_xay f_y f_cond) /\
            den_gt0_ninfinity f_xay f_y f_cond ==>
            (Rel p (R_CSP Y X) t =
             1 -
             pos_fn_integral lborel
               (\y.
                    indicator_fn {u | 0 <= u /\ u <= t} y * f_y y *
                    pos_fn_integral lborel
                      (\x. indicator_fn {w | y < w /\ w <= t} x * f_cond y x)))``,

RW_TAC std_ss[Rel_def, o_DEF]
THEN `prob_space p` by FULL_SIMP_TAC std_ss[cond_density_def, random_variable_def]
THEN KNOW_TAC(``DRBD_event p (R_CSP Y X) t = p_space p DIFF (DFT_event p( CSP Y X) t) ``)
     THEN1(RW_TAC std_ss[DRBD_event_def, DFT_event_def, CSP_def, R_CSP_def]
           THEN SRW_TAC[][GSPECIFICATION, EXTENSION]
 	   THEN EQ_TAC THEN METIS_TAC[extreal_lt_def, extreal_le_def, le_total, lt_total])

THEN RW_TAC std_ss[]
THEN KNOW_TAC(``DFT_event p (CSP Y X) t IN events p ``)
     THEN1(FULL_SIMP_TAC std_ss[rv_gt0_ninfinity_def] THEN RW_TAC std_ss[ DFT_CSP_event_GT0]
           THEN RW_TAC std_ss[PREIMAGE_EXTREAL_REAL_2RV_BEFORE_CSP]
	   THEN FULL_SIMP_TAC std_ss[cond_density_def]
	   THEN KNOW_TAC(`` (\x. real(X x),real(Y x)) IN
          measurable (m_space p,measurable_sets p)
            (m_space (pair_measure lborel lborel),
             measurable_sets (pair_measure lborel lborel)) ``)
	       THEN1(KNOW_TAC(`` random_variable (\s. real (X s)) p
	                            (m_space lborel, measurable_sets lborel)  ``)
	            THEN1(RW_TAC std_ss[])
		    THEN KNOW_TAC(`` random_variable (\s. real (Y s)) p
		                    (m_space lborel, measurable_sets lborel)  ``)
	            THEN1(RW_TAC std_ss[])
		    THEN RW_TAC std_ss[]
		    THEN METIS_TAC[measurable_Pair, p_space_def, random_variable_def, events_def, space_def, m_space_def, prob_space_def, subsets_def,prob_space_def, measure_space_lborel])
	  THEN RW_TAC std_ss[]	    
          THEN METIS_TAC[IN_MEASURABLE, CSP_set_IN_MEASURABLE_SETS_PAIR_lborel, p_space_def, events_def, subsets_def, measurable_sets_def, m_space_def, space_def])
THEN RW_TAC std_ss[]	  
THEN RW_TAC std_ss[PROB_COMPL, EVENTS_COMPL]
THEN METIS_TAC[prob_CSP]);



val Rel_R_WSP = store_thm("Rel_R_WSP",
  `` !p Y X_a X_d t f_y f_xay f_cond.
            0<=t /\ (!s. ALL_DISTINCT [X_a s; X_d s; Y s]) /\
            DISJOINT_WSP Y X_a X_d t/\ rv_gt0_ninfinity [X_a; X_d; Y] /\
           den_gt0_ninfinity f_xay f_y f_cond /\
            
            (!y.
                 cond_density lborel lborel p (real o X_a)
                   (real o Y) y f_xay f_y f_cond) /\
            indep_var p lborel (real o X_d) lborel (real o Y) /\
            cont_CDF p (real o X_d) /\
            measurable_CDF p (real o X_d) ==>
            (Rel p (R_WSP Y X_a X_d) t =
             1 -
             (pos_fn_integral lborel
                (\y.
                     indicator_fn {u | 0 <= u /\ u <= t} y * f_y y *
                     pos_fn_integral lborel
                       (\x. indicator_fn {w | y < w /\ w <= t} x * f_cond y x)) +
              pos_fn_integral lborel
                (\y.
                     f_y y *
                     (indicator_fn {u | 0 <= u /\ u <= t} y *
                      CDF p (real o X_d) y)))) ``,

RW_TAC std_ss[Rel_def, o_DEF]
THEN `prob_space p` by FULL_SIMP_TAC std_ss[indep_var, indep_vars, indep_sets]
THEN KNOW_TAC(`` DRBD_event p (R_WSP Y X_a X_d) t =
                 p_space p DIFF( DFT_event p (WSP Y X_a X_d) t)``)

     THEN1(RW_TAC std_ss[DRBD_event_def, DFT_event_def, R_WSP_def1, WSP_def]
           THEN RW_TAC std_ss[R_AND_def,R_AFTER_def, D_SIMULT_def, D_BEFORE_def,D_OR_def, D_AND_def]
	   THEN FULL_SIMP_TAC std_ss[rv_gt0_ninfinity_def, ALL_DISTINCT, MEM]THEN SRW_TAC[][GSPECIFICATION, EXTENSION] THEN EQ_TAC THEN RW_TAC std_ss[extreal_min_def, extreal_max_def] THEN REPEAT( POP_ASSUM MP_TAC) THEN RW_TAC std_ss[] THEN FULL_SIMP_TAC std_ss[]THEN   METIS_TAC[extreal_min_def, extreal_max_def, extreal_lt_def, extreal_le_def, lt_le, le_lt, lt_total, le_total, lt_infty, le_infty, rv_gt0_ninfinity_def])
THEN RW_TAC std_ss[]
THEN KNOW_TAC(``DFT_event p (WSP Y X_a X_d) t IN events p ``)
     THEN1(KNOW_TAC(``WSP Y X_a X_d =
                      D_OR (D_AND (X_a)(D_BEFORE Y X_a))(D_AND (Y)(D_BEFORE X_d Y)) ``)   
           THEN1(FULL_SIMP_TAC std_ss[R_OR_def, R_NEVER_def, rv_gt0_ninfinity_def, ALL_DISTINCT, MEM]
	         THEN RW_TAC std_ss[D_OR_def, D_AND_def,D_BEFORE_def, WSP_def, D_SIMULT_def]
		 THEN KNOW_TAC(``
    !s.
       (min
          (min
             (min (max (X_a s) (if Y s < X_a s then Y s else PosInf))
                (max (Y s) (if X_d s < Y s then X_d s else PosInf)))
             (if Y s = X_a s then X_a s else PosInf))
          (if Y s = X_d s then X_d s else PosInf)) =

       (min (max (X_a s) (if Y s < X_a s then Y s else PosInf))
          (max (Y s) (if X_d s < Y s then X_d s else PosInf)))``)
	         THEN1( RW_TAC std_ss[extreal_max_def, extreal_min_def] THEN REPEAT (POP_ASSUM MP_TAC) THEN RW_TAC std_ss[] THEN FULL_SIMP_TAC std_ss[] THEN METIS_TAC[lt_infty, le_infty, lt_le, le_lt, lt_trans, le_trans, extreal_lt_def, extreal_le_def, ALL_DISTINCT, MEM])
		 THEN RW_TAC std_ss[])
	 THEN DISCH_TAC THEN ONCE_ASM_REWRITE_TAC[]	 
         THEN RW_TAC std_ss[OR_UNION]
	 THEN KNOW_TAC(`` DFT_event p (D_AND X_a (D_BEFORE Y X_a)) t IN events p ``)
	     THEN1(`rv_gt0_ninfinity [Y; X_a]` by FULL_SIMP_TAC std_ss[rv_gt0_ninfinity_def]
	            THEN RW_TAC std_ss[DFT_event_AFTER_PREIMAGE_GT0, rv_gt0_ninfinity_def]
	          THEN `{(u,w) | u < w /\ 0 <= w /\ w <= t} IN
		        measurable_sets (pair_measure lborel lborel)` by RW_TAC std_ss[after_set_IN_MEASURABLE_SETS_PAIR_lborel]
		  THEN FULL_SIMP_TAC std_ss[cond_density_def]
                  THEN `(\x. (real (Y x),real (X_a x))) IN
		        measurable (m_space p, measurable_sets p) (m_space (pair_measure lborel lborel), measurable_sets (pair_measure lborel lborel))`
		       by METIS_TAC[random_variable_def, p_space_def, events_def, m_space_def, measurable_sets_def, measurable_Pair, prob_space_def,  measure_space_lborel, borel, lborel, subsets_def, space_def, sigma_def]
		  THEN METIS_TAC[IN_MEASURABLE, random_variable_def, space_def, m_space_def, subsets_def, measurable_sets_def, events_def, p_space_def])
        THEN DISCH_TAC
	THEN KNOW_TAC(`` DFT_event p (D_AND Y (D_BEFORE X_d Y)) t IN events p``)
	 THEN1(`rv_gt0_ninfinity [X_d; Y]` by FULL_SIMP_TAC std_ss[rv_gt0_ninfinity_def]
	        THEN RW_TAC std_ss[DFT_event_AFTER_PREIMAGE_GT0, rv_gt0_ninfinity_def]
		THEN FULL_SIMP_TAC std_ss[cond_density_def]
	          THEN `{(u,w) | u < w /\ 0 <= w /\ w <= t} IN measurable_sets (pair_measure lborel lborel)` by RW_TAC std_ss[after_set_IN_MEASURABLE_SETS_PAIR_lborel]
                  THEN `(\x. (real (X_d x),real (Y x))) IN measurable (m_space p, measurable_sets p) (m_space (pair_measure lborel lborel), measurable_sets (pair_measure lborel lborel))`
		       by METIS_TAC[random_variable_def, p_space_def, events_def, m_space_def, measurable_sets_def, measurable_Pair, prob_space_def,  measure_space_lborel, borel, lborel, subsets_def, space_def, sigma_def, indep_var_rv]
		  THEN METIS_TAC[IN_MEASURABLE, random_variable_def, space_def, m_space_def, subsets_def, measurable_sets_def, events_def, p_space_def, indep_var_rv])
        THEN DISCH_TAC
	THEN METIS_TAC [EVENTS_UNION])
THEN RW_TAC std_ss[PROB_COMPL, EVENTS_COMPL]
THEN FULL_SIMP_TAC std_ss[R_OR_def, R_NEVER_def]
THEN DEP_REWRITE_TAC[prob_WSP]
THEN Q.EXISTS_TAC`f_xay`
THEN RW_TAC std_ss[prob_WSP, R_OR_def, R_NEVER_def, D_AND_def, NEVER_def]);


val FDEP_R_AND = store_thm("FDEP_R_AND",
  `` !X Y p t. DFT_event p (FDEP X Y) t = p_space p DIFF (DRBD_event p (R_AND X Y) t)  ``,

RW_TAC std_ss[DFT_event_def, DRBD_event_def, R_AND_def, FDEP_def]
THEN SRW_TAC [] [GSPECIFICATION, EXTENSION]
THEN EQ_TAC THEN RW_TAC std_ss[extreal_min_def, extreal_max_def] THEN FULL_SIMP_TAC std_ss[] THEN METIS_TAC [extreal_lt_def, extreal_le_def, lt_le, le_lt, lt_total, le_total]);


val PAND_R_AFTER = store_thm("PAND_R_AFTER",
  `` 
        !X Y p t.
            (!s. ALL_DISTINCT [X s; Y s]) ==>
            (DFT_event p (P_AND Y X) t =
             p_space p DIFF DRBD_event p (R_AFTER X Y) t) ``,

RW_TAC std_ss[DFT_event_def, DRBD_event_def, P_AND_def, R_AFTER_def]
THEN FULL_SIMP_TAC std_ss[ALL_DISTINCT, MEM]
THEN SRW_TAC [] [GSPECIFICATION, EXTENSION] THEN EQ_TAC THEN RW_TAC std_ss[] THEN  METIS_TAC [extreal_lt_def, extreal_le_def, lt_le, le_lt, lt_total, le_total, lt_infty, le_infty, ALL_DISTINCT, MEM]);



val R_AFTER_PAND = store_thm("R_AFTER_PAND",
  ``!X Y p t.
        (!s.  ALL_DISTINCT [X s;Y s]) ==>
	(DRBD_event p (R_AFTER X Y) t =
	 p_space p DIFF (DFT_event p (P_AND Y X) t))  ``,

RW_TAC std_ss[DFT_event_def, DRBD_event_def, P_AND_def, R_AFTER_def]
THEN FULL_SIMP_TAC std_ss[ALL_DISTINCT, MEM]
THEN SRW_TAC [] [GSPECIFICATION, EXTENSION] THEN EQ_TAC THEN RW_TAC std_ss[] THEN  METIS_TAC [extreal_lt_def, extreal_le_def, lt_le, le_lt, lt_total, le_total, lt_infty, le_infty, ALL_DISTINCT, MEM]);



val R_AFTER_D_AFTER = store_thm("R_AFTER_D_AFTER",
  ``!X Y p t.
         (DRBD_event p (R_AFTER X Y) t =
	  p_space p DIFF (DFT_event p (D_AND X(D_BEFORE Y X) )t))  ``,

RW_TAC std_ss[DFT_event_def, DRBD_event_def, D_BEFORE_def, D_AND_def, R_AFTER_def]
THEN FULL_SIMP_TAC std_ss[ALL_DISTINCT, MEM]
THEN SRW_TAC [] [GSPECIFICATION, EXTENSION] THEN EQ_TAC THEN RW_TAC std_ss[] THEN  METIS_TAC [extreal_lt_def, extreal_le_def, lt_le, le_lt, lt_total, le_total, lt_infty, le_infty, ALL_DISTINCT, MEM, extreal_max_def]);



val PROB_R_AFTER = store_thm("PROB_R_AFTER",
  ``   !X Y p fx t.
            rv_gt0_ninfinity [X; Y] /\ 0 <= t /\ prob_space p /\
            indep_var p lborel (\s. real (X s)) lborel (\s. real (Y s)) /\
            distributed p lborel (\s. real (X s)) fx /\ (!x. 0 <= fx x) /\
            cont_CDF p (\s. real (Y s)) /\ measurable_CDF p (\s. real (Y s)) ==>
            (prob p (DRBD_event p (R_AFTER X Y) t) =
             1 -
             pos_fn_integral lborel
               (\x.
                    fx x *
                    (indicator_fn {u | 0 <= u /\ u <= t} x *
                     CDF p (\s. real (Y s)) x)))``,
RW_TAC std_ss[R_AFTER_D_AFTER]
THEN KNOW_TAC(``DFT_event p (D_AND X (D_BEFORE Y X)) t IN events p ``)
     THEN1(`rv_gt0_ninfinity[Y;X]` by FULL_SIMP_TAC std_ss[rv_gt0_ninfinity_def]
           THEN RW_TAC std_ss[DFT_event_AFTER_PREIMAGE_GT0]
           THEN `(\x. (real (Y x),real (X x))) IN measurable (m_space p, measurable_sets p) (m_space (pair_measure lborel lborel), measurable_sets (pair_measure lborel lborel))`
		       by METIS_TAC[random_variable_def, p_space_def, events_def, m_space_def, measurable_sets_def, measurable_Pair, prob_space_def,  measure_space_lborel, borel, lborel, subsets_def, space_def, sigma_def, indep_var_rv]
		  THEN METIS_TAC[IN_MEASURABLE, random_variable_def, space_def, m_space_def, subsets_def, measurable_sets_def, events_def, p_space_def,after_set_IN_MEASURABLE_SETS_PAIR_lborel])
THEN RW_TAC std_ss[PROB_COMPL, EVENTS_COMPL]
THEN `rv_gt0_ninfinity[Y;X]` by FULL_SIMP_TAC std_ss[rv_gt0_ninfinity_def]
THEN FULL_SIMP_TAC std_ss[indep_var]
THEN KNOW_TAC(`` indep_var p lborel (\s. real (Y s)) lborel (\s. real (X s)) ``)           
     THEN1( MP_TAC(Q.ISPECL[`(\i:num. if i = 0 then (\s. real (X s)) else (\s. real ((Y:'a->extreal) s)))`,`p: 
        ('a -> bool) #
        (('a -> bool) -> bool) # (('a -> bool) -> extreal)`,`lborel`, `1:num`, `0:num`, `{x | (x = 0) \/ (x = 1:num)}` ] indep_vars_var)
	   THEN PSET_TAC[])
THEN METIS_TAC[PROB_DFT_AFTER]);


val Rel_R_AFTER = store_thm("Rel_R_AFTER",
  ``  !X Y p fx t.
           rv_gt0_ninfinity [X; Y] /\ 0 <= t  /\
          indep_var p lborel ( real o X) lborel (real o Y) /\
          distributed p lborel (real o X) fx /\ (!x. 0 <= fx x) /\
          cont_CDF p (real o Y) /\ measurable_CDF p (real o Y) ==>
          (Rel p (R_AFTER X Y) t =
           1 - pos_fn_integral lborel
             (\x.
                  fx x *
                  (indicator_fn {u | 0 <= u /\ u <= t} x *
                   CDF p (real o Y) x)))``,
RW_TAC std_ss[Rel_def, o_DEF]
THEN `prob_space p` by FULL_SIMP_TAC std_ss[indep_var, indep_vars, indep_sets]
THEN RW_TAC std_ss[PROB_R_AFTER]);


(******************************************************************************)
(*                                                                            *)
(*                             series-parallel-series                         *)
(*                                as a special case of                        *)
(*                              nested series-parallel                        *)
(******************************************************************************)

val BIGUNION_o_BIGUNION_def = Define `BIGUNION_o_BIGUNION s L A =
                                      BIGUNION {s l | l IN BIGUNION {L a | a IN A }} `;


val DRBD_series_parallel_series_lem = store_thm(
  "DRBD_series_parallel_series_lem",
  `` !p X L A J.
            DRBD_series
              (\j.
                   DRBD_parallel
                     (\a. DRBD_series (\l. DRBD_parallel X {l}) (L a)) (A j))
              J =
            DRBD_series (\j. DRBD_parallel (\a. DRBD_series X (L a)) (A j)) J``,

RW_TAC std_ss[]
THEN RW_TAC std_ss[DRBD_series_def, DRBD_parallel_def]
THEN FULL_SIMP_TAC std_ss[BIGUNION_o_BIGUNION_def]
THEN AP_TERM_TAC
THEN RW_TAC std_ss[] THEN SRW_TAC[][GSPECIFICATION, EXTENSION]
THEN EQ_TAC THEN RW_TAC std_ss[] THEN Q.EXISTS_TAC`j` THEN RW_TAC std_ss[] THEN EQ_TAC
THEN RW_TAC std_ss[] THEN  Q.EXISTS_TAC`s` THEN RW_TAC std_ss[] THEN Q.EXISTS_TAC`a`
THEN  ASM_SET_TAC[]);



val DRBD_series_parallel_series_lem2 = store_thm(
  "DRBD_series_parallel_series_lem2",
  ``!p X s L A J.
            (!l. l IN BIGUNION_o_BIGUNION L A J ==> (s l = {l})) ==>
            (DRBD_series
               (\j.
                    DRBD_parallel
                      (\a. DRBD_series (\l. DRBD_parallel X (s l)) (L a))
                      (A j)) J =
             DRBD_series (\j. DRBD_parallel (\a. DRBD_series X (L a)) (A j))
               J)``,


RW_TAC std_ss[]
THEN RW_TAC std_ss[DRBD_series_def, DRBD_parallel_def]
THEN FULL_SIMP_TAC std_ss[BIGUNION_o_BIGUNION_def]
THEN AP_TERM_TAC
THEN RW_TAC std_ss[] THEN SRW_TAC[][GSPECIFICATION, EXTENSION]
THEN EQ_TAC THEN RW_TAC std_ss[] THEN Q.EXISTS_TAC`j` THEN RW_TAC std_ss[] THEN EQ_TAC
THEN RW_TAC std_ss[] THEN  Q.EXISTS_TAC`s'` THEN RW_TAC std_ss[] THEN Q.EXISTS_TAC`a`
THEN  ASM_SET_TAC[]);


val PROB_DRBD_series_parallel_series_lem1 = store_thm(
  "PROB_DRBD_series_parallel_series_lem1",
  `` !p X s L A J.
            (!l. l IN BIGUNION_o_BIGUNION L A J ==> (s l = {l})) /\
            indep_sets p (\i. {X i}) (BIGUNION_o_BIGUNION L A J) /\
            sets_finite_not_empty s L A J ==>
            (prob p
               (DRBD_series
                  (\j.
                       DRBD_parallel
                         (\a. DRBD_series (\l. DRBD_parallel X (s l)) (L a))
                         (A j)) J) =
             Normal
               (product J
                  (\j.
                       real
                         (1 -
                          Normal
                            (product (A j)
                               (\a.
                                    real
                                      (1 -
                                       Normal
                                         (product (L a)
                                            (\l.
                                                 real
                                                   (1 -
                                                    Normal
                                                      (real
                                                         (1 - prob p (X l)))))))))))))``,

RW_TAC std_ss[]
THEN KNOW_TAC(`` (!l:num. l IN BIGUNION_o_BIGUNION (L:num->num->bool) (A:num->num->bool) (J:num->bool) ==> (X:num->'a->bool) l IN events p)  ``)
     THEN1(FULL_SIMP_TAC std_ss[indep_sets]
           THEN ASM_SET_TAC[])
THEN RW_TAC std_ss[]	   
THEN DEP_REWRITE_TAC[PROB_DRBD_nested_series_parallel]
THEN RW_TAC std_ss[]
     THEN1(FULL_SIMP_TAC std_ss[nested_BIGUNION_def, BIGUNION_o_BIGUNION_def] 
           THEN KNOW_TAC(``BIGUNION
	                      {s (l:num) | l IN BIGUNION
			                          {L (a:num) | a IN BIGUNION
						                     {A (j:num) | j IN J}}} =
                           BIGUNION {L (a:num) | a IN BIGUNION {A (j:num) | j IN J}}  ``)
                THEN1(ASM_SET_TAC[])
	    THEN DISCH_TAC THEN FULL_SIMP_TAC std_ss[])

THEN KNOW_TAC(``
         !j.
	  j IN (J:num->bool) ==>
	   (real
            (1 -
             Normal
               (product ((A:num->num->bool) j)
                  (\a.
                       real
                         (1 -
                          Normal
                            (product ((L:num->num->bool) a)
                               (\l.
                                    real
                                      (1 -
                                       Normal
                                         (product ((s:num->num->bool) l)
                                            (\i. real (1 - prob p ((X:num->'a->bool) i))))))))))) =

          real
            (1 -
             Normal
               (product (A j)
                  (\a.
                       real
                         (1 -
                          Normal
                            (product (L a)
                               (\l.
                                    real
                                      (1 - Normal (real (1 - prob p (X l))))))))))) ``)
      THEN1(RW_TAC std_ss[]
           THEN NTAC 3 (AP_TERM_TAC)
	   THEN KNOW_TAC(``
	   !a.
	    a IN (A:num->num->bool) j==> (  real
            (1 -
             Normal
               (product ((L:num->num->bool) a)
                  (\l.
                       real
                         (1 -
                          Normal
                            (product ((s:num->num->bool) l) (\i. real (1 - prob p ((X:num->'a->bool) i))))))))   =
	    real
            (1 -
             Normal
               (product (L a)
                  (\l. real (1 - Normal (real (1 - prob p (X l))))))))``)
		THEN1(RW_TAC std_ss[]
		     THEN NTAC 3(AP_TERM_TAC)
		     THEN KNOW_TAC(``
		         !l:num.
			    l IN (L:num->num->bool) a ==>
		            (real (1 - Normal (product ((s:num->num->bool) l) (\i. real (1 - prob p ((X:num->'a->bool) i))))) =
			    real (1 - Normal (real (1 - prob p (X l)))) )``)
			 THEN1(RW_TAC std_ss[]
			       THEN `(s:num->num->bool) (l:num) = {l}` by ASM_SET_TAC[BIGUNION_o_BIGUNION_def]
			       THEN ONCE_ASM_REWRITE_TAC[]
			       THEN NTAC 3(AP_TERM_TAC)
			       THEN RW_TAC std_ss[PRODUCT_SING])
		     THEN RW_TAC std_ss[th])
	   THEN RW_TAC std_ss[th])
THEN RW_TAC std_ss[th]);


val PROB_DRBD_series_parallel_series = store_thm(
  "PROB_DRBD_series_parallel_series",
  ``  !p X s L A J.
            (!l. l IN BIGUNION_o_BIGUNION L A J ==> (s l = {l})) /\
            indep_sets p (\i. {X i}) (BIGUNION_o_BIGUNION L A J) /\
            sets_finite_not_empty s L A J ==>
            (prob p
               (DRBD_series
                  (\j. DRBD_parallel (\a. DRBD_series X (L a)) (A j)) J) =
             Normal
               (product J
                  (\j.
                       real
                         (1 -
                          Normal
                            (product (A j)
                               (\a.
                                    real
                                      (1 -
                                       Normal
                                         (product (L a)
                                            (\l.
                                                 real
                                                   (1 -
                                                    Normal
                                                      (real
                                                         (1 - prob p (X l)))))))))))))``,
							 
RW_TAC std_ss[]
THEN KNOW_TAC(``
       DRBD_series 
        (\j. DRBD_parallel
	      (\a. DRBD_series (X:num->'a->bool) ((L:num->num->bool) a)) ((A:num->num->bool) j)) (J:num->bool) =
	     DRBD_series 
              (\j.
                  DRBD_parallel 
                    (\a.
                         DRBD_series
			    (\l. DRBD_parallel  X ((s:num->num->bool) l))
                          (L a)) (A j)) J ``)
     THEN1 RW_TAC std_ss[GSYM DRBD_series_parallel_series_lem2]
THEN DISCH_TAC THEN ONCE_ASM_REWRITE_TAC[]     
THEN RW_TAC std_ss[PROB_DRBD_series_parallel_series_lem1]);


val sets_finite_not_empty2_def = Define `sets_finite_not_empty2 L A J=
          (!i. i IN J ==> A i <> {} /\ FINITE (A i)) /\
          (!i. i IN BIGUNION {A j | j IN J} ==> L i <> {} /\ FINITE (L i)) /\
          FINITE J /\ disjoint_family_on A J /\ J <> {} /\
          disjoint_family_on L (BIGUNION {A j | j IN J})`;


val sets_finite_not_empty_empty2 = store_thm("sets_finite_not_empty_empty2",
  `` !L A J.
            sets_finite_not_empty2 L A J <=>
            sets_finite_not_empty (\l. {l}) L A J``, 

RW_TAC std_ss[sets_finite_not_empty2_def, sets_finite_not_empty_def]
THEN EQ_TAC
THEN RW_TAC std_ss[]
     THEN1(ASM_SET_TAC[])
     THEN1(ASM_SET_TAC[FINITE_SING])
THEN RW_TAC std_ss[disjoint_family_on]     
THEN ASM_SET_TAC[]);     


val PROB_DRBD_series_parallel_series_lem = store_thm(
  "PROB_DRBD_series_parallel_series_lem",
  `` !p X s L A J.
            indep_sets p (\i. {X i}) (BIGUNION_o_BIGUNION L A J) /\
            sets_finite_not_empty2 L A J ==>
            (prob p
               (DRBD_series
                  (\j. DRBD_parallel (\a. DRBD_series X (L a)) (A j)) J) =
             Normal
               (product J
                  (\j.
                       real
                         (1 -
                          Normal
                            (product (A j)
                               (\a.
                                    real
                                      (1 -
                                       Normal
                                         (product (L a)
                                            (\l.
                                                 real
                                                   (1 -
                                                    Normal
                                                      (real
                                                         (1 - prob p (X l)))))))))))))``,

RW_TAC std_ss[]
THEN MP_TAC(Q.ISPECL [`p:('a->bool)#(('a->bool)->bool)#(('a->bool)->extreal)`,`X:num->'a->bool`,`(\l:num. {l})`,`L:num->num->bool`,`A:num->num->bool`,`J:num->bool`] PROB_DRBD_series_parallel_series_lem1)
THEN RW_TAC std_ss[]
THEN `sets_finite_not_empty(\l.{l}) L A J` by METIS_TAC[sets_finite_not_empty_empty2]
THEN FULL_SIMP_TAC std_ss[]
THEN METIS_TAC[DRBD_series_parallel_series_lem]);


val PROB_DRBD_series_parallel_series_rv_red    = store_thm(
  "PROB_DRBD_series_parallel_series_rv_red",
  `` !p X s L A J t.
            indep_sets p (\i. {rv_to_event p X t i}) (BIGUNION_o_BIGUNION L A J) /\
            sets_finite_not_empty2 L A J ==>
            (prob p
               (DRBD_series
                  (\j. DRBD_parallel (\a. DRBD_series (rv_to_event p X t) (L a)) (A j)) J) =
             Normal
               (product J
                  (\j.
                      1 -
                         product (A j)
                               (\a.
                                   1 -
                                       product (L a)
                                            (\l.
                                                1 -
                                                   real
                                                         (1 - Rel p (X l) t))))))``,

RW_TAC std_ss[PROB_DRBD_series_parallel_series_lem, Rel_def, rv_to_event_def]
THEN RW_TAC std_ss[ normal_real_product_1, real_normal_real, normal_real_product_2]);

val _ = export_theory();


